/*
EL.h Written by Walter O. Krawec
Header file for Evolution List (EL) Learning

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef MTENN_EL_H_
#define MTENN_EL_H_

#include "STM.h"

#include <vector>
#include <list>

#ifdef MTENN_STM_DEBUG
#include <iostream>
#endif

namespace MTENN
{
	struct LearningSet
	{
		LearningSet() : r(0), lifeTime(-1) {};

		std::vector <float> inputs, outputs;
		float r;
		int lifeTime;

		bool checkInputsEqual(const LearningSet& set)
		{
			if(inputs.size() != set.inputs.size())
				return false;

			const float e = 0.01f;
			for(unsigned int i=0; i<inputs.size(); ++i)
			{
				if(fabs(inputs[i] - set.inputs[i]) > e)
					return false;
			}

			return true;
		}

		bool operator==(const LearningSet& set)
		{
			if(inputs.size() != set.inputs.size() || outputs.size() != set.outputs.size())
				return false;

			const float e = 0.01f;
			for(unsigned int i=0; i<inputs.size(); ++i)
			{
				if(fabs(inputs[i] - set.inputs[i]) > e)
					return false;
			}
			for(unsigned int i=0; i<outputs.size(); ++i)
			{
				if(fabs(outputs[i] - set.outputs[i]) > e)
					return false;
			}

			return true;
		}

		void print()
		{
#ifdef MTENN_DEBUG
			std::cout << "(<";
			for(unsigned int i=0; i<inputs.size(); ++i)
				std::cout << inputs[i] << ",";
			std::cout << ">, <";
			for(unsigned int i=0; i<outputs.size(); ++i)
				std::cout << outputs[i] << ",";
			std::cout << ">, " << r << ", " << lifeTime << ")";
#endif
		}
	};

	class EvolutionList : public STM
	{
	public:
		EvolutionList() : binaryMode(false) {};

		bool elementInList_byInput(LearningSet& set)
		{
			if(learningList.empty())
				return false;

			std::list <LearningSet>::iterator Iter;
			for(Iter = learningList.begin(); Iter != learningList.end(); ++Iter)
			{
				if(set.checkInputsEqual(*Iter)) return true;
			}

			return false;
		}

		void print()
		{
#ifdef MTENN_DEBUG
			std::cout << "Evolution List = {";
			if(learningList.empty())
			{
				std::cout << "EMPTY}\n";
				return;
			}

			std::list <LearningSet>::iterator Iter;
			for(Iter = learningList.begin(); Iter != learningList.end(); ++Iter)
			{
				Iter->print();
				std::cout << ",\n";
			}
			std::cout << "}\n";
#endif
		}
		void addToListDirect(LearningSet& l)
		{
			learningList.push_back(l);
			if(learningList.size() > 20)
				learningList.erase(learningList.begin());
		}
		void clearList()
		{
			learningList.clear();
		}
		void addToList(LearningSet& l)
		{
			if(learningList.empty())
			{
				learningList.push_back(l);
				return;
			}

			std::list <LearningSet>::iterator Iter;
			for(Iter = learningList.begin(); Iter != learningList.end(); ++Iter)
			{
				if(l == *Iter)
				{
					Iter->r += l.r;
					//Iter->lifeTime += l.lifeTime;
					if(l.lifeTime > Iter->lifeTime)
						Iter->lifeTime = l.lifeTime;
					return;
				}
			}

			learningList.push_back(l);
		}
		void multiplyScalar(float c)
		{
			if(learningList.empty())
				return;

			std::list <LearningSet>::iterator Iter;
			for(Iter = learningList.begin(); Iter != learningList.end(); ++Iter)
				Iter->r *= c;
		}
		std::list <LearningSet>& getLearningList() {return learningList;}
		void mainLoop()
		{
			if(!learningList.empty())
			{
				std::list<LearningSet>::iterator Iter;
				Iter = learningList.begin();
				while(Iter != learningList.end())
				{
					if(Iter->lifeTime == 0 || fabs(Iter->r) < 0.1f)
					{
						learningList.erase(Iter);
						if(learningList.empty())
							break;
						Iter = learningList.begin();
					}
					else
					{
						if(Iter->lifeTime > 0)
						{
							Iter->lifeTime -= 0;//10;
							if(Iter->lifeTime < 0)
								Iter->lifeTime = 0;
						}
						++Iter;
					}
				}
			}
		}

		float fitness(ENN* network);
		float maxFitness();
		float minFitness();

		void degrade(float scalar)
		{
			if(!learningList.empty())
			{
				std::list <LearningSet>::iterator Iter;
				for(Iter = learningList.begin(); Iter != learningList.end(); ++Iter)
					Iter->r *= scalar;
			}
		}

		void addToMemory(const std::vector <float>& inputs, const std::vector <float>& outputs, float reward)
		{
			LearningSet set;
			set.inputs = inputs;
			set.outputs = outputs;
			if(binaryMode)
			{
				for(unsigned int i=0; i<set.outputs.size(); ++i)
				{
					if(set.outputs[i] >= 0.5f)
						set.outputs[i] = 1;
					else
						set.outputs[i] = 0;
				}
			}
			set.r = reward;
			set.lifeTime = 1;

			addToList(set);
		}

		bool resize(int numInputs, int numOutputs)
		{
			return true;
		}

		bool setup(int numInputs, int numOutputs)
		{
			return true;
		}

		void setBinaryOutputMode(bool b) {binaryMode = b;}
	private:
		std::list <LearningSet> learningList;

		bool binaryMode;
	};
}

#endif