/*
ENN.cpp
Written by Walter O. Krawec
Implementation of Evolving Neural Networks

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "ENN.h"

namespace MTENN
{
	void Neuron::run(ENN* network)
	{
		if(!network)
			return;

		float sum=0;

		for(unsigned int i=0; i<inputs.size(); ++i)
		{
			if(inputs[i].valid)
			{
				sum += inputs[i].get();

				if(sum > 10000)
				{
					int error = 1;
					error *= 1;
				}
			}
		}

		sum += localValue;

		sum = activation(sum);
		if(sum > 1000)
			sum = 1000.0f;
		else if(sum < -1000)
			sum = -1000.0f;

		lastOutput = sum;

		if(!outputs.empty())
		{
			std::list <Output>::iterator Iter;
			for(Iter = outputs.begin(); Iter != outputs.end(); ++Iter)
			{
				if(Iter->valid)
					network->setNeuronInput(Iter->toNeuron, Iter->toInput, sum);
			}
		}
	}

	void ENN::addIO(unsigned int numIn, unsigned int numOut)
	{
		numInputs += numIn;
		numOutputs += numOut;
		for(unsigned int i=0; i<numIn; ++i)
		{
			Neuron n(Neuron::Linear, true);
			unsigned int index=neuronList.size();
			for(unsigned int i=0; i<neuronList.size(); ++i)
			{
				if(!neuronList[i].isValid())
				{
					index = i;
					break;
				}
			}

			if(index == neuronList.size())
				neuronList.push_back(n);
			else
				neuronList[index] = n;

			inputs.push_back(index);

			for(unsigned int j=0; j<outputs.size(); ++j)
			{
				Neuron::Input input(index);
				unsigned int toNeuron = outputs[j];
				neuronList[index].outputs.push_back(Neuron::Output(toNeuron, neuronList[toNeuron].addInput(input)));
			}
		}

		for(unsigned int i=0; i<numOut; ++i)
		{
			Neuron n(Neuron::Sigmoid, true);
			unsigned int index=neuronList.size();
			for(unsigned int i=0; i<neuronList.size(); ++i)
			{
				if(!neuronList[i].isValid())
				{
					index = i;
					break;
				}
			}

			if(index == neuronList.size())
				neuronList.push_back(n);
			else
				neuronList[index] = n;

			outputs.push_back(index);
		}
	}
	bool ENN::create(unsigned int numIn, unsigned int numOut, unsigned int numRandom, ConnectionType connection)
	{
		cleanup();
		Neuron n;
		neuronList.resize(numIn+numOut+numRandom);

		// first add neuron for each IO and random:

		for(unsigned int i=0; i<numIn+numOut; ++i)
		{
			if(i < numIn)
				n = Neuron(Neuron::Linear, true);
			else
				n = Neuron(Neuron::Sigmoid, true);
			neuronList[i] = n;

			if(i < numIn)
				lastEvolveTape.evolve_addNeuron(EvolveTape::Input);
			else
				lastEvolveTape.evolve_addNeuron(EvolveTape::Output);
		}

		// now add input pointer:
		for(unsigned int i=0; i<numIn; ++i)
			inputs.push_back(i);

		// and output pointer:
		for(unsigned int i=0; i<numOut; ++i)
			outputs.push_back(i+numIn);

		numInputs = numIn;
		numOutputs = numOut;

		if(connection == NoConnection)
			return true;

		if(connection == FF)
		{
			// and connect them (simple FF):
			for(unsigned int i=0; i<numIn; ++i)
			{
				for(unsigned int j=0; j<numOut; ++j)
				{
					Neuron::Input input(i);
					neuronList[i].outputs.push_back(Neuron::Output(j+numIn, neuronList[j+numIn].addInput(input)));
					lastEvolveTape.evolve_addSynapse(i, j+numIn, input.weight, input.value);
				}
			}
		}
		else if(connection == Hemisphere)
		{
			// and connect them
			for(unsigned int i=0; i<numIn/2; ++i)
			{
				for(unsigned int j=0; j<numOut/2; ++j)
				{
					Neuron::Input input(i);
					neuronList[i].outputs.push_back(Neuron::Output(j+numIn, neuronList[j+numIn].addInput(input)));
					lastEvolveTape.evolve_addSynapse(i, j+numIn, input.weight, input.value);
				}
			}

			for(unsigned int i=numIn/2; i<numIn; ++i)
			{
				for(unsigned int j=numOut/2; j<numOut; ++j)
				{
					Neuron::Input input(i);
					neuronList[i].outputs.push_back(Neuron::Output(j+numIn, neuronList[j+numIn].addInput(input)));
					lastEvolveTape.evolve_addSynapse(i, j+numIn, input.weight, input.value);
				}
			}
		}

		for(unsigned int i=0; i<numRandom; ++i)
			addRandomNeuron();

		clearLastEvolveTape(true);

		return true;
	}

	unsigned int ENN::chooseRandomNeuron(bool avoidIO)
	{
		if(neuronList.size() == 0)
			return 0;

		unsigned int index = rand()%neuronList.size(), check=index, stop;
		if(index == 0)
			stop = neuronList.size();
		else
			stop = index-1;

		while(check != stop)
		{
			if(check >= neuronList.size())
				check = 0;
			if(neuronList[check].isValid())
			{
				if(!avoidIO)
					return check;
				else if(neuronList[check].isIO())
					return check;
			}
			++check;
		}

		return neuronList.size();	// couldn't find a neuron
	}

	void ENN::setInput(unsigned int index, float value)
	{
		if(index >= inputs.size() || inputs[index] >= neuronList.size() || !neuronList[inputs[index]].isValid())//!(*inputs[index]))
			return;

		neuronList[inputs[index]].setLocalValue(value);
	}
	float ENN::getInput(unsigned int index)
	{
		if(index >= inputs.size() || inputs[index] >= neuronList.size() || !neuronList[inputs[index]].isValid())//!(*inputs[index]))
			return 0;

		return neuronList[inputs[index]].getLocalValue();
	}
	float ENN::getOutput(unsigned int index)
	{
		if(index >= outputs.size() || outputs[index] >= neuronList.size() || !neuronList[outputs[index]].isValid())// !(*outputs[index]))
			return 0;

		return neuronList[outputs[index]].getLastOutput();
	}
	void ENN::run()
	{
		for(unsigned int i=0; i<inputs.size(); ++i)
		{
			if(neuronList[inputs[i]].isValid())
				neuronList[inputs[i]].run(this);
		}
		for(unsigned int i=0; i<neuronList.size(); ++i)
		{
			if(neuronList[i].isIO()) continue;

			if(neuronList[i].isValid())
				neuronList[i].run(this);
		}

		for(unsigned int i=0; i<outputs.size(); ++i)
		{
			if(neuronList[outputs[i]].isValid())
				neuronList[outputs[i]].run(this);
		}
	}

	bool ENN::synapseExists(unsigned int fromNeuron, unsigned int toNeuron)
	{
		if(toNeuron >= neuronList.size() || !neuronList[toNeuron].isValid())
			return false;
		if(fromNeuron >= neuronList.size() || !neuronList[fromNeuron].isValid())
			return false;

		if(neuronList[fromNeuron].outputs.empty())
			return false;

		std::list <Neuron::Output>::iterator Iter;
		for(Iter = neuronList[fromNeuron].outputs.begin(); Iter != neuronList[fromNeuron].outputs.end(); ++Iter)
		{
			if(Iter->toNeuron == toNeuron)
				return true;
		}

		return false;
	}

	void ENN::evolveFromTape(EvolveTape& tape)
	{
		evolveTape.append(tape);
		tape.resetPositionPointer();
		//for(unsigned int i=0; i<tape.size(); ++i)
		while(1)
		{
			int type = evolveTape.getType();
			if(type == EvolveTape::endOfTape)
				break;

			if(type == EvolveTape::eolMarker) continue;
			else if(type == EvolveTape::addNeuron)
			{
				// add neuron
				EvolveTape::NeuronData data;
				tape.getNeuronBlock(data);

				Neuron n(Neuron::Sigmoid, false);

				if(data.neuronType == EvolveTape::Input)
				{
					n.type = Neuron::Linear;
					n.IONeuron = true;
				}
				else if(data.neuronType == EvolveTape::Output)
					n.IONeuron = true;

				unsigned int index=neuronList.size();
				for(unsigned int j=0; j<neuronList.size(); ++j)
				{
					if(!neuronList[j].isValid())
					{
						index = j;
						break;
					}
				}

				if(index == neuronList.size())
					neuronList.push_back(n);
				else
					neuronList[index] = n;
			}
			else if(type == EvolveTape::addSynapse)
			{
				// add synapse:
				EvolveTape::SynapseData data;
				tape.getSynapseBlock(data);

				unsigned int from = data.from;
				unsigned int to = data.to;
				float weight = data.weight;
				float value = data.value;

				if(to >= neuronList.size() || from >= neuronList.size())
				{
					std::cout << "Error\n";
					continue;
				}

				if(synapseExists(from, to) || !neuronList[to].isValid() || !neuronList[from].isValid())
					continue;

				Neuron::Input input(from);
				input.weight = weight;
				input.value = value;

				neuronList[from].addOutput(Neuron::Output(to, neuronList[to].addInput(input)));
			}
			else if(type == EvolveTape::removeNeuron)
			{
				// remove neuron
				EvolveTape::NeuronData data;
				tape.getNeuronBlock(data);
				unsigned int id = data.index;
				removeNeuron(id);
			}
			else if(type == EvolveTape::removeSynapse)
			{
				// remove synapse
				EvolveTape::SynapseData data;
				tape.getSynapseBlock(data);
				unsigned int from = data.from;
				unsigned int to = data.to;

				removeSynapse(from, to);
			}
			else if(type == EvolveTape::setWeight)
			{
				// change weight
				EvolveTape::SynapseData data;
				tape.getSynapseBlock(data);
				unsigned int from = data.from;
				unsigned int to = data.to;
				float weight = data.weight;

				changeWeight(from, to, weight);
			}
			else
				throw("ENN::evolveFromTape() Error: block type not defined");
		}
	}

	void ENN::evolve(STM* stm, int level)
	{
		const int MAX_RECURSION_LEVEL		= 2;
		const int LEVEL1_NUMBER_OF_NETWORKS	= 100;
		const int DIV_FACTOR				= 8;
		const int SCALE_TO_CONSIDER			= 10;

		if(!stm)
			return;

		// add check to see if stm is empty?

		if(level == 1)
			stm->mainLoop();

		if(level > MAX_RECURSION_LEVEL)
			return;

		
		const int N = LEVEL1_NUMBER_OF_NETWORKS/(1+DIV_FACTOR*(level-1));	// number of evolutions to try

		std::list <ENN> evolutionList;
		computeLocalFitness(stm);


		if(level == 1)
			clearLastEvolveTape(true);

		evolutionList.push_back(*this);
		
		for(int i=1; i<=N; ++i)
		{
			ENN tempNetwork = *this;
			tempNetwork.randomEvolve();
			tempNetwork.computeLocalFitness(stm);
			evolutionList.push_back(tempNetwork);
		}

		evolutionList.sort();
		
		const int K=N/SCALE_TO_CONSIDER;

		std::list <ENN>::iterator Iter=evolutionList.end();
		--Iter;
		for(int j=0; j<K; ++j)
			(*Iter).evolve(stm, level+1);

		for(int j=0; j<K/2; ++j)
		{
			int index = rand()%N;
			for(Iter = evolutionList.begin(); Iter != evolutionList.end(); ++Iter)
			{
				--index;
				if(index < 0)
					break;
			}

			(*Iter).evolve(stm, level+1);
		}

		evolutionList.sort();

		Iter = evolutionList.end();
		--Iter;
		*this = *Iter;

		return;
	}

	void ENN::randomEvolve()
	{
		//addRandomSynapse();
		int p = rand()%200;

		if(neuronEvolveCounter > 0)
		{
			modifyWeights();
			--neuronEvolveCounter;
			return;
		}

		if(p < 2)
			addRandomNeuron(true);
		else if(p < 4)
			removeRandomNeuron();
		else if(p < 7)
			addRandomSynapse();
		else if(p < 10)
			removeRandomSynapse();
		else
			modifyWeights();
	}

	void ENN::addRandomNeuron(bool setWaitTimer)
	{
		Neuron n(Neuron::Sigmoid, false);

		unsigned int index=neuronList.size();
		for(unsigned int i=0; i<neuronList.size(); ++i)
		{
			if(!neuronList[i].isValid())
			{
				index = i;
				break;
			}
		}

		if(index == neuronList.size())
			neuronList.push_back(n);
		else
			neuronList[index] = n;

		lastEvolveTape.evolve_addNeuron(EvolveTape::Hidden);

		// now connect it:
		unsigned int index1 = index;
		while(index1 == index)
			index1 = chooseRandomNeuron();
		unsigned int index2 = index;
		while(index2 == index)
			index2 = chooseRandomNeuron();
		/*
		if(index1 == index || index2 == index)
		{
			std::cout << "ERROR.\n";
			return;
		}*/

		if(index1 < neuronList.size() && neuronList[index1].isValid())
		{
			Neuron::Input input(index);
			neuronList[index].addOutput(Neuron::Output(index1, neuronList[index1].addInput(input)));
			lastEvolveTape.evolve_addSynapse(index, index1, input.weight, input.value);
		}
		if(index2 < neuronList.size() && neuronList[index2].isValid())
		{
			Neuron::Input input(index2);
			neuronList[index2].addOutput(Neuron::Output(index, neuronList[index].addInput(input)));
			lastEvolveTape.evolve_addSynapse(index2, index, input.weight, input.value);
		}

		if(setWaitTimer)
			neuronEvolveCounter = NeuronEvolveWaitTime;
	}

	void ENN::removeRandomNeuron(bool setWaitTime)
	{
		if(neuronList.empty())
			return;

		unsigned int k = rand()%neuronList.size();

		if(!neuronList[k].isValid() || neuronList[k].isIO())
			return;

		removeNeuron(k);
		lastEvolveTape.evolve_removeNeuron(k);

		if(setWaitTime)
			neuronEvolveCounter = NeuronEvolveWaitTime;
	}

	void ENN::addRandomSynapse(bool setWaitTime)
	{
		if(neuronList.empty())
			return;

		unsigned int to = rand()%neuronList.size();
		unsigned int from = rand()%neuronList.size();

		if(synapseExists(from, to) || !neuronList[to].isValid() || !neuronList[from].isValid())
			return;

		Neuron::Input input(from);

		neuronList[from].addOutput(Neuron::Output(to, neuronList[to].addInput(input)));

		lastEvolveTape.evolve_addSynapse(from, to, input.weight, input.value);

		if(setWaitTime)
			synapseEvolveCounter = SynapseEvolveWaitTime;
	}


	void ENN::changeWeight(unsigned int from, unsigned int to, float weight)
	{
		if(to >= neuronList.size())
		{
			std::cout << "Error: to = " << to << ", size = " << neuronList.size() << "\n";
			return;
		}
		for(unsigned int i=0; i<neuronList[to].inputs.size(); ++i)
		{
			if(neuronList[to].inputs[i].from == from)
				neuronList[to].inputs[i].weight = weight;
		}
	}

	void ENN::removeSynapse(unsigned int from, unsigned int to)
	{
		bool found = false;
		int r = 0;
		std::list < Neuron::Output >::iterator Iter;
		for(Iter = neuronList[from].outputs.begin(); Iter != neuronList[from].outputs.end(); ++Iter)
		{
			if(Iter->toNeuron == to)
			{
				r = Iter->toInput;
				found = true;
				neuronList[from].outputs.erase(Iter);
				break;
			}
		}

		if(!found)
		{
			std::cout << "ERROR - Cannot find correct Output link.\n";
			return;
		}

		neuronList[to].inputs[r].valid = false;
	}

	void ENN::removeRandomSynapse(bool setWaitTime)
	{
		if(neuronList.size() == 0)
			return;

		unsigned int k = rand()%neuronList.size();
		if(!neuronList[k].isValid())
			return;

		if(!neuronList[k].inputs.empty())
		{
			unsigned int r = rand()%(neuronList[k].inputs.size());
			if(!neuronList[k].inputs[r].valid)
				return;

			unsigned int fromNeuron = neuronList[k].inputs[r].from;
			if(fromNeuron >= neuronList.size() || neuronList[fromNeuron].outputs.empty())
				return;							// Error?

			bool found = false;
			std::list < Neuron::Output >::iterator Iter;
			for(Iter = neuronList[fromNeuron].outputs.begin(); Iter != neuronList[fromNeuron].outputs.end(); ++Iter)
			{
				if(Iter->toNeuron == k && Iter->toInput == r)
				{
					found = true;
					neuronList[fromNeuron].outputs.erase(Iter);
					break;
				}
			}

			if(!found)
			{
				std::cout << "ERROR - Cannot find correct Output link.\n";
				return;
			}

			neuronList[k].inputs[r].valid = false;

			lastEvolveTape.evolve_removeSynapse(fromNeuron, k);
		}

		if(setWaitTime)
			synapseEvolveCounter = SynapseEvolveWaitTime;
	}

	void ENN::modifyWeights()
	{
		if(neuronList.empty())
			return;

		const float deltaW = 0.04f;

		for(unsigned int i=0; i<neuronList.size(); ++i)
		{
			if(!neuronList[i].isValid() || neuronList[i].inputs.empty())
				continue;

			for(unsigned int j=0; j<neuronList[i].inputs.size(); ++j)
			{
				if(!neuronList[i].inputs[j].valid)
					continue;

				if(rand()%100 < 15)
				{
					neuronList[i].inputs[j].weight = Utilities::randFloat(-10,10);//+= Utilities::randFloat(-10*deltaW, 10*deltaW);
					lastEvolveTape.evolve_setWeight(neuronList[i].inputs[j].from, i, neuronList[i].inputs[j].weight);
				}
			}
		}
		return;
	}

	void ENN::removeNeuron(unsigned int index)
	{
		if(index >= neuronList.size() || !neuronList[index].isValid())
			return;

		if(!neuronList[index].inputs.empty())
		{
			for(unsigned int i=0; i<neuronList[index].inputs.size(); ++i)
			{
				unsigned int from = neuronList[index].inputs[i].from;
				if(from >= neuronList.size() || neuronList[from].outputs.empty())
					continue;
				std::list <Neuron::Output>::iterator Iter;
				for(Iter = neuronList[from].outputs.begin(); Iter != neuronList[from].outputs.end(); ++Iter)
				{
					if(Iter->toNeuron == index && Iter->toInput == i)
					{
						neuronList[from].outputs.erase(Iter);
						break;
					}
				}
			}
		}

		if(!neuronList[index].outputs.empty())
		{
			std::list <Neuron::Output>::iterator Iter;
			for(Iter = neuronList[index].outputs.begin(); Iter != neuronList[index].outputs.end(); ++Iter)
			{
				if(Iter->toNeuron < neuronList.size() && Iter->toInput < neuronList[Iter->toNeuron].inputs.size())
					neuronList[Iter->toNeuron].inputs[Iter->toInput].valid = false;
			}
		}

		neuronList[index].valid = false;
	}
}