/*
ENN.h Written by Walter O. Krawec
Primary header file for Evolving Neural Networks

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef MTENN_H_
#define MTENN_H_

// define MTENN_DEBUG and/or MTENN_STM_DEBUG to get debug information (printed to console)
// in your main program (before including ENN.h, MTENN.h, ULearn.h, and EL.h
// Careful: MTENN_STM_DEBUG will print out a lot of stuff!
//
// also make sure you seed the random number generator before use

#include "WOK\Tape.h"

#include <list>
#include <vector>
#include <algorithm>

#ifdef MTENN_DEBUG
#include <iostream>
#endif

#include "EvolveTape.h"
#include "Util.h"
#include "STM.h"

namespace MTENN
{
	class ENN;

	class Neuron
	{
		friend class ENN;
	public:
		struct Input
		{
			Input()
			{
				const float minValue = -4.0f;
				const float maxValue =  4.0f;
				weight = Utilities::randFloat(minValue, maxValue);
				value = Utilities::randFloat();

				from = 0;

				valid = true;
			}

			Input(unsigned int _from)
			{
				const float minValue = -4.0f;
				const float maxValue =  4.0f;
				weight = Utilities::randFloat(minValue, maxValue);
				value = Utilities::randFloat();

				from = _from;

				valid = true;
			}

			float get()
			{
				return weight*value;
			}

			float weight, value;
			unsigned int from;

			bool valid;
		};
		struct Output
		{
			Output() : toNeuron(0), toInput(0), valid(true) {};
			Output(unsigned int _toNeuron, unsigned int _toInput) : toNeuron(_toNeuron), toInput(_toInput), valid(true) {};

			unsigned int toNeuron, toInput;
			bool valid;
		};
		enum Type
		{
			Sigmoid = 0,
			Linear,
			Threshold
		};

		Neuron()
		{
			type = Sigmoid;
			IONeuron = false;
			lastOutput = 0;
			localValue = 0;
			valid = false;
		}
		Neuron(Type t, bool isIONeuron) : type(t), IONeuron(isIONeuron)
		{
			lastOutput = 0;
			localValue = 0;
			valid = true;
		};

		void setValid(bool value) {valid = value;};
		bool isValid() {return valid;};

		void run(ENN* network);

		void addOutput(Output o)
		{
			outputs.push_back(o);
		}

		bool isIO() {return IONeuron;};

		void setLocalValue(float value) {localValue = value;}
		float getLocalValue() {return localValue;}
		float getLastOutput() {return lastOutput;}

		unsigned int addInput(Input input)
		{
			unsigned int index = inputs.size();
			for(unsigned int i=0; i<inputs.size(); ++i)
			{
				if(!inputs[i].valid)
				{
					index = i;
					break;
				}
			}

			if(index >= inputs.size())
				inputs.push_back(input);
			else
				inputs[index] = input;

			return index;
		}
	private:
		float activation(float x)
		{
			const float t = 0.5;
			switch(type) {
				case Sigmoid: 
							  if(x > 80.0f) return 1.0f;
							  else if(x < -80.0f) return 0.0f;
							  else return (1.0f / (1.0f + exp(-x)));
							  break;
				case Linear: return x; break;
				case Threshold: return (x > t)?1.0f:0.0f; break;
			}

			return 0;
		}

		std::vector <Input> inputs;
		std::list <Output> outputs;

		bool IONeuron;

		float localValue, lastOutput;

		Type type;

		bool valid;
	};

	class ENN
	{
		friend class Neuron;
	public:
		enum ConnectionType
		{
			NoConnection=0,
			FF,
			Hemisphere
		};

		ENN()
		{
			numInputs = 0;
			numOutputs = 0;

			NeuronEvolveWaitTime = 40;
			SynapseEvolveWaitTime = 25;

			neuronEvolveCounter = 0;
			synapseEvolveCounter = 0;

			localFitness = 0;
		};
		~ENN()
		{
			cleanup();
		}

		bool operator< (const ENN& e)
		{
			return (localFitness < e.localFitness) ? true : false;
		}

		void cleanup()
		{
			neuronList.clear();
			inputs.clear();
			outputs.clear();

			evolveTape.clear();
			lastEvolveTape.clear();
		}

		bool create(unsigned int numIn, unsigned int numOut, unsigned int numRandom, ConnectionType connection);
		void addIO(unsigned int numIn, unsigned int numOut);
		void randomEvolve();

		void evolve(STM* stm, int level=1);

		void evolveFromTape(EvolveTape& tape);
		void evolveFromTape()
		{
			evolveFromTape(evolveTape);
		}

		float fitness(STM* stm) {if(stm) return stm->fitness(this); return 0;};

		float getLocalFitness() {return localFitness;}

		void setInput(unsigned int index, float value);
		float getInput(unsigned int index);
		int getInputInt(unsigned int index)
		{
			return (getInput(index) >= 0.5) ? 1 : 0;
		}

		float getOutput(unsigned int index);
		int getOutputInt(unsigned int index)
		{
			return (getOutput(index) >= 0.5) ? 1 : 0;
		}

		unsigned int getNumOutputs()
		{
			return outputs.size();
		}
		unsigned int getNumInputs()
		{
			return inputs.size();
		}

		void run();

		EvolveTape& getLastEvolveTape() {return lastEvolveTape;}

		void printDebugStatus()
		{
			std::cout << "ENN Debug Status\n";
			std::cout << "Number of Neurons = " << neuronList.size() << "\n";
			std::cout << "EvolveTape Size = " << evolveTape.size() << "\n";
		}

		float getNeuronValue(unsigned int index)
		{
			if(index >= neuronList.size())
				return 0.0f;
			if(neuronList[index].isValid())
				return neuronList[index].getLastOutput();
			return 0;
		}
	private:
		void setNeuronInput(unsigned int neuronIndex, unsigned int inputIndex, float value)
		{
			if(neuronIndex >= neuronList.size() || !neuronList[neuronIndex].isValid())
				return;
			if(inputIndex >= neuronList[neuronIndex].inputs.size() || !neuronList[neuronIndex].inputs[inputIndex].valid)
				return;

			neuronList[neuronIndex].inputs[inputIndex].value = value;
		}

		void computeLocalFitness(STM* stm) {if(stm) localFitness = stm->fitness(this);};

		void removeNeuron(unsigned int index);
		void removeSynapse(unsigned int from, unsigned int to);

		void changeWeight(unsigned int from, unsigned int to, float weight);

		void clearLastEvolveTape(bool addToEvolveTape=true)
		{
			if(addToEvolveTape)
				evolveTape.append(lastEvolveTape);
			lastEvolveTape.clear();
		}
		

		unsigned int chooseRandomNeuron(bool avoidIO=false);

		void addRandomNeuron(bool setWaitTime=false);
		void removeRandomNeuron(bool setWaitTime=false);
		void addRandomSynapse(bool setWaitTime=false);
		void removeRandomSynapse(bool setWaitTime=false);
		void modifyWeights();

		bool synapseExists(unsigned int fromNeuron, unsigned int toNeuron);		// returns TRUE if there is already a synapse between "to" and "from"

		std::vector <Neuron> neuronList;
		std::vector < unsigned int > inputs, outputs;
		unsigned int numInputs, numOutputs;

		float localFitness;

		int neuronEvolveCounter;		// number of cycles left until we may remove a neuron
		int NeuronEvolveWaitTime;		// number of cycles to keep a new neuron

		int synapseEvolveCounter;
		int SynapseEvolveWaitTime;

		EvolveTape evolveTape, lastEvolveTape;
	};
}

#endif