/*
ENNInterface.h Written by Walter O. Krawec
A helpful interface for using an ENN
Highly recommended to use this interface instead of "manually" interfacing to an ENN

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

/*
Directions for Use:
1) First create your instincts by inheriting MTENN::ENNInterface::Instinct
   a) One of which should be a "Boredom" instinct (see "simulation.h")
      In particular, it should produce a "correct" trigger if robot has been
	  moving around for a certain amount of time; an "incorrect" trigger if
	  the robot has been sitting still for too long. The "correct" trigger
	  is important for the DP system to function correctly.  Unlike what is hinted
	  at in the article, this interface does not manage the "correct" paths
	  automatically.  It was done in this isolated fashion to avoid timing issues.
2) Define your reflexes by inheriting from MTENN::ENNInterface::Reflex
3) Decide which STM to use (e.g. U-Learn or EL-Learn) and declare a memory object
   a) Example: MTENN::ULearn U; // if using U-Learning
   b)          MTENN::EvolutionList EL; // if using EL-Learning
4) Define an ENNInterface:
   a) Example: MTENN::ENNInterface::Interface ai;
   b) Call ai.setup(...);
5) Define instances of each instinct and reflex you use
6) Add reflexes to the correct instincts (via INSTINCT.addReflex(&REFLEX))
7) Add Instincts to Interface (via INTERFACE.addInstinct(&INSTINCT))
   a) See setupRobot() for examples
8) Run the ENN
   a) First setup using ai.setup(...) (see setupRobot())
   b) Repeat the following (see mainLoop()):
      i)   Use ai.setInput(INDEX, VALUE)
      ii)  ai.run()
      iii) ai.getOuput(INDEX)
      iv)  ai.mainLoop();	// call this to run DP manager and also to evolve on occasion.
*/

#ifndef MTENN_INTERFACE_H_
#define MTENN_INTERFACE_H_

#include <vector>
#include <list>

#include "STM.h"
#include "ENN.h"

#ifdef MTENN_DEBUG
#include <iostream>
#endif

namespace MTENN
{
namespace ENNInterface
{
	class Interface;

	class Reflex
	{
	public:
		virtual void trigger()=0;
	};

	class Instinct
	{
		friend class Interface;
	public:
		// isTriggered is polled by Interface::mainLoop() and should return true if the instinct has triggered
		// Once triggered, this should continue to return true until "reset" is called regardless of whether or not the instinct
		// is triggered.

		virtual bool isTriggered()=0;

		// isCorrect returns true if this instinct determines a "correct" response (i.e. a reward); else it is incorrect (i.e. a penalty)
		virtual bool isCorrect()=0;

		// reset is called after isTriggered() returns true.  May not be immediate however isTriggered() should continue to report "true" (if triggered once)
		// until reset is called
		virtual void reset()=0;

		// returns the scale factor used in the decision path for this instinct.
		virtual float getDPScaleFactor(){return 0.5f;};

		// returns the reward value for this instinct (negative for penalty)
		virtual float getReward()=0;

		//virtual float getCurrentScaleFactor()=0;

		// used for debuggin - should print the name of this instinct to the screen (if MTENN_DEBUG is defined)
		virtual void printName() {};

		// reflex manager - at setup, add reflexes using "addReflex".  When this instinct is triggered, all added reflexes
		// are run (Reflex::trigger())
		void addReflex(Reflex* r) {reflexList.push_back(r);};

	private:
		void triggerReflex()
		{
			if(reflexList.empty())
				return;

			std::list <Reflex*>::iterator Iter;
			for(Iter = reflexList.begin(); Iter != reflexList.end(); ++Iter)
			{
				if(!(*Iter)) continue;
				(*Iter)->trigger();
			}
		}
		std::list <Reflex*> reflexList;
	};

	class DPManager
	{
		struct Decision
		{
			Decision() {};
			Decision(int numInputs, int numOutputs)
			{
				input.resize(numInputs);
				output.resize(numOutputs);
			}

			std::vector <float> input, output;
		};
	public:
		DPManager() : points(NULL), stm(NULL), maxPoints(0), currentIndex(0) {};
		~DPManager() {cleanup();}

		void cleanup()
		{
			if(points)
				delete[] points;

			points = NULL;
			maxPoints = 0;
			currentIndex = 0;

			stm = NULL;
		}

		bool setup(int _maxPoints, int numIn, int numOut, STM* _stm)
		{
			if(!_stm)
				return false;

			cleanup();

			stm = _stm;
			maxPoints = _maxPoints;
			points = new Decision[maxPoints];
			currentIndex = 0;
			
			stm->setup(numIn, numOut);

			inputs.clear();
			for(int i=0; i<numIn; ++i)
				inputs.push_back(0);

			outputs.clear();
			for(int i=0; i<numOut; ++i)
				outputs.push_back(0);

			return true;
		}

		void resize(int numIn, int numOut)
		{
			if(!stm)
				return;
			stm->resize(numIn, numOut);

			inputs.clear();
			outputs.clear();

			for(int i=0; i<numIn; ++i)
				inputs.push_back(0);
			for(int i=0; i<numOut; ++i)
				outputs.push_back(0);
		}

		void addSpecificToSTM(const std::vector <float>& in, const std::vector <float>& out, float reward)
		{
			if(!stm)
				return;
			stm->addToMemory(in, out, reward);
		}

		void addCurrentToSecondarySTM(ENN* network)
		{
			if(!network)
				return;

			if(inputs.size() != network->getNumInputs() || outputs.size() != network->getNumOutputs())
			{
#ifdef MTENN_DEBUG
				std::cout << "DPManager::addCurrentToSecondarySTM() Error - IO vectors wrong size.\n";
#endif
				return;
			}

			for(unsigned int i=0; i<network->getNumInputs(); ++i)
				inputs[i] = network->getInput(i);
			for(unsigned int i=0; i<network->getNumOutputs(); ++i)
				outputs[i] = network->getOutput(i);

#ifdef MTENN_DEBUG
			std::cout << "Adding <";
			for(unsigned int i=0; i<inputs.size(); ++i)
				std::cout << inputs[i] << ", ";
			std::cout << "> -> <";
			for(unsigned int i=0; i<outputs.size(); ++i)
				std::cout << outputs[i] << ", ";
			std::cout << ">; " << "\n";//(int)correct << "\n";
#endif
			points[currentIndex].input = inputs;
			points[currentIndex].output = outputs;
			++currentIndex;
			if(currentIndex >= maxPoints)
			{
				overflow=true;
				currentIndex = 0;
			}
		}

		void addPathToSTM(bool correct, float scaleFactor, float reward)
		{
			if(!stm)
				return;
			float c = 1;
			if(!correct)
				c = -1;

			float scale = 1;

			for(int i=currentIndex-1; i >= 0; --i)
			{
				stm->addToMemory(points[i].input, points[i].output, reward*scale*c);
				scale = scale*scaleFactor;
			}
			if(overflow)
			{
				for(int i=maxPoints-1; i>=currentIndex; --i)
				{
					stm->addToMemory(points[i].input, points[i].output, reward*scale*c);
					scale = scale*scaleFactor;
				}
			}

			currentIndex = 0;
			overflow = false;
		}

		STM* getSTM() {return stm;};
	private:
		int maxPoints;
		Decision* points;
		std::vector <float> inputs, outputs;
		STM* stm;
		int currentIndex;
		bool overflow;
	};

	// Primary Interface to an ENN system (including STM and Decision Paths)
	class Interface
	{
	public:
		// Future work.
		enum DPMode
		{
			Timed=0,
			Change
		};

		Interface() : timer(0), DPTime(100), evolveTime(100), pauseEvolve(false) {};

		// call setup first; point it to the STM system you want to use (e.g. U-Learning or EL Learning)
		// numInputs, etc. are clear
		// _DPMaxSecondaryPoints specifies the max. size of a decision path to keep in memory before rolling over and erasing past events
		// DPTime is the time (in ai.mainLoop() iterations) between DP points
		// evolveTime is the number of ai.mainLoop() iterations between calls to network.evolve()
		void setup(STM* _stm, int numInputs, int numOutputs, int numHidden, int _DPMaxSecondaryPoints, int _DPTime, int _evolveTime)
		{
			network.create(numInputs, numOutputs, numHidden, ENN::Hemisphere);
			DPTime = _DPTime;
			evolveTime = _evolveTime;
			timer = 0;

			dp.setup(_DPMaxSecondaryPoints, numInputs, numOutputs, _stm);

			pauseEvolve = false;
		}

		void setPauseEvolve(bool value) {pauseEvolve = value;}

		// call these procedures to use the network
		void setInput(int index, float value)
		{
			network.setInput(index, value);
		}
		float getOutput(int index)
		{
			return network.getOutput(index);
		}
		int getOutputInt(int index)
		{
			return network.getOutputInt(index);
		}
		void runNetwork()
		{
			network.run();
		}

		ENN& getNetwork() {return network;};
		DPManager& getDP() {return dp;};

		// call this in your main loop to keep things running
		void mainLoop(bool evolveOnly=false)
		{
			++timer;
			if(!evolveOnly)
			{
				if(timer % DPTime == 0)
					dp.addCurrentToSecondarySTM(&network);

				if(!instincts.empty())
				{
					std::list <Instinct*>::iterator Iter;
					for(Iter = instincts.begin(); Iter != instincts.end(); ++Iter)
					{
						if(!(*Iter))
							continue;

						if((*Iter)->isTriggered())
						{
#ifdef MTENN_DEBUG
							std::cout << "Instinct Triggered: ";
							(*Iter)->printName();
							std::cout << "\n";
#endif
							dp.addCurrentToSecondarySTM(&network);
							dp.addPathToSTM((*Iter)->isCorrect(), (*Iter)->getDPScaleFactor(), (*Iter)->getReward());

							(*Iter)->triggerReflex();

							(*Iter)->reset();
						}
					}
				}
			}

			if(timer % evolveTime == 0)
			{
				evolve(2);
				if(dp.getSTM())
					dp.getSTM()->degrade(0.8f);
			}
		}

		void evolve(int numIterations)
		{
			if(pauseEvolve)
				return;

			for(int i=0; i<numIterations; ++i)
				network.evolve(dp.getSTM());
		}

		void addInstinct(Instinct* i)
		{
			instincts.push_back(i);
		}

		// This sort of works; it should report a number between 0 and 1 with 0 being "most unfit" and 1 being complete fit
		// w.r.t. the current STM state.  Works for U-Learning not EL learning.
		// could be useful in the future to scale back the time to evolve if a network has reached its goal (again w.r.t STM state)
		float getTotalFitness(bool recomputeFitness=false)
		{
			if(!dp.getSTM())
				return 0.0f;

			
			float fit = network.getLocalFitness();
			if(recomputeFitness)
				fit = dp.getSTM()->fitness(&network);
			float maxFit = dp.getSTM()->maxFitness();
			float minFit = dp.getSTM()->minFitness();

			if(fabs(minFit - maxFit) < 0.001f)
				return 1.0f;

			float m = 1.0f / (maxFit - minFit);

			fit = (fit-minFit)*m;

			if(fit < 0)
				return 0;
			if(fit > 1)
				return 1;

			return fit;
		}
	private:
		DPManager dp;
		ENN network;
		std::list <Instinct*> instincts;
		int DPTime, evolveTime;
		int timer;
		bool pauseEvolve;
	};
}
}

#endif