/*
EvolveTape.cpp
Written by Walter O. Krawec
Implementation of an ENN's evolve tape

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "EvolveTape.h"

namespace MTENN
{
	int EvolveTape::getType()
	{
		if(positionPointer >= (int)tape.size())
			return endOfTape;

		while(tape.get(positionPointer) == eolMarker)
		{
			++positionPointer;
			if(positionPointer >= (int)tape.size())
				return endOfTape;
		}
		return tape[positionPointer];
	}

	bool EvolveTape::getSynapseBlock(SynapseData& output)
	{
		int type = getType();
		if(type != addSynapse || type != removeSynapse || type != setWeight)
			return false;

		output.from = tape.get(positionPointer+1);
		output.to = tape.get(positionPointer+2);

		switch(type)
		{
		case addSynapse:
			output.weight = Utilities::IntToFloat( tape.get(positionPointer+3) );
			output.value = Utilities::IntToFloat( tape.get(positionPointer+4) );
			positionPointer += 5;
			break;

		case setWeight:
			output.weight = Utilities::IntToFloat( tape.get(positionPointer+3) );
			positionPointer += 4;
			break;
		}

		return true;
	}

	bool EvolveTape::getNeuronBlock(NeuronData& output)
	{
		int type = getType();
		if(type != addNeuron || type != removeNeuron)
			return false;

		if(type == removeNeuron)
			output.index = tape.get(positionPointer+1);
		else
			output.neuronType = tape.get(positionPointer+1);

		positionPointer += 2;

		return true;
	}

	void EvolveTape::append(EvolveTape& tape2)
	{
		tape.append(tape2.getTape());
	}

	bool EvolveTape::saveTapeToFile(const std::string& filename)
	{
		std::ofstream f;
		f.open(filename.c_str());
		if(!f.is_open())
			return false;

		for(unsigned int i=0; i<tape.size(); ++i)
			f << (int)tape[i] << "\n";

		f << 0 << "\n";
		f.close();

		return true;
	}

	bool EvolveTape::loadTapeFromFile(const std::string& filename)
	{
		std::ifstream f;
		f.open(filename.c_str());
		if(!f.is_open())
			return false;

		char buffer[100];
		while(!f.eof())
		{
			f.getline(buffer, 100);
			tape.push(atoi(buffer));
		}
		f.close();
		return true;
	}

	void EvolveTape::printTapeNice(std::ostream& f, wok::Tape <int>& t)
	{
		std::ofstream f2;
		f2.open("test.txt");
		for(unsigned int i=0; i<t.size(); ++i)
		{
			f2 << (int) t[i] << ", ";
		}
		for(unsigned int i=0; i<t.size(); ++i)
		{
			if(t[i] == -1) continue;
			else if(t[i] == 1)
				f << "Add Neuron\n";
			else if(t[i] == 2)
			{
				f << "Add Synapse: ";
				f << t[i+1] << " -> " << t[i+2];
				f << "; weight = " << Utilities::IntToFloat(t[i+3]) << ", value = " << Utilities::IntToFloat(t[i+4]);
				f << "\n";
				i+=4;
			}
			else if(t[i] == 3)
			{
				f << "Remove Neuron; Index = " << t[i+1] << "\n";
				++i;
			}
			else if(t[i] == 4)
			{
				f << "Remove Synapse: ";
				f << t[i+1] << " -> " << t[i+2] << "\n";
				i += 2;
			}
			else if(t[i] == 5)
			{
				f << "Set Synapse Weight: ";
				f << t[i+1] << " -> " << t[i+2];
				f << "; weight = " << Utilities::IntToFloat(t[i+3]) << "\n";
				i += 3;
			}
			else
				f << "\nEvolve Tape Error\n\n";
		}

		f2.close();
	}
}