/*
EvolveTape.h Written by Walter O. Krawec
Manages the Evolution Tape system of an ENN

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef MTENN_EVOLVE_TAPE_H_
#define MTENN_EVOLVE_TAPE_H_

#include "WOK\Tape.h"
#include <fstream>

#include "Util.h"

namespace MTENN
{
	class EvolveTape
	{
	public:
		enum
		{
			DEFAULT_SINGLE_TAPE_SIZE= 2048,			// controls how large a single tape section is
													// when this is filled, a new tape of this size is created

			MAX_DATA_PER_BLOCK		= 4,			// the largest amount of data per tape block (used by TapeBlock)

			endOfTape				= -2,
			eolMarker				= -1,
			addNeuron				= 1,
			addSynapse				= 2,
			removeNeuron			= 3,
			removeSynapse			= 4,
			setWeight				= 5
		};

		enum NeuronType
		{
			Input = 0,
			Output,
			Hidden
		};

		struct SynapseData
		{
			SynapseData() : from(0), to(0), weight(0), value(0) {};

			unsigned int from, to;
			float weight, value;
		};
		struct NeuronData
		{
			NeuronData() : index(0) {}
			unsigned int index;
			int neuronType;
		};

		EvolveTape() : tape(DEFAULT_SINGLE_TAPE_SIZE)
		{
			positionPointer = 0;
			addEOL = true;
		}

		bool saveTapeToFile(const std::string& filename);
		bool loadTapeFromFile(const std::string& filename);
		void printTapeNice(std::ostream& f, wok::Tape <int>& t);

		// to read an evolve tape (easily) first reset the position pointer
		// next use getType to determine block type
		// then use getSynapseBlock or getNeuronBlock as needed.
		// repeat until getType() return endOfTape
		void resetPositionPointer() {positionPointer=0;}
		int  getType();
		bool getSynapseBlock(SynapseData& output);
		bool getNeuronBlock(NeuronData& output);

		void clear()
		{
			tape.clear();
		}
		void append(EvolveTape& tape2);

		unsigned int size() {return tape.size();};

		// these are called by ENN::evolve()
		void evolve_addNeuron(NeuronType type)
		{
			tape.push(addNeuron);
			tape.push(type);

			if(addEOL)
				tape.push(eolMarker);
		}
		void evolve_addSynapse(unsigned int from, unsigned int to, float weight, float value)
		{
			tape.push(addSynapse);
			tape.push(from);
			tape.push(to);
			tape.push(Utilities::FloatToInt(weight));
			tape.push(Utilities::FloatToInt(value));

			if(addEOL)
				tape.push(eolMarker);
		}
		void evolve_removeNeuron(unsigned int index)
		{
			tape.push(removeNeuron);
			tape.push((int)index);

			if(addEOL)
				tape.push(eolMarker);
		}
		void evolve_removeSynapse(unsigned int from, unsigned int to)
		{
			tape.push(removeSynapse);
			tape.push((int)from);
			tape.push((int)to);

			if(addEOL)
				tape.push(eolMarker);
		}
		void evolve_setWeight(unsigned int from, unsigned int to, float weight)
		{
			tape.push(setWeight);
			tape.push((int)from);
			tape.push((int)to);
			tape.push(Utilities::FloatToInt(weight));

			if(addEOL)
				tape.push(eolMarker);
		}

		wok::Tape <int>& getTape() {return tape;}
	private:
		wok::Tape <int> tape;
		int positionPointer;
		bool addEOL;
	};
};

#endif