/*
ULearn.h Written by Walter O. Krawec
Header file for U-Learning class; inherits class STM.

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef MTENN_U_LEARN_H_
#define MTENN_U_LEARN_H_

#include <vector>
#include <fstream>

#include "STM.h"

namespace MTENN
{
	class Matrix
	{
	public:
		Matrix() : m(NULL), size(0), columns(0), rows(0) {};
		Matrix(int r, int c)
		{
			m = NULL;
			size = 0;
			create(r,c);
		}
		~Matrix() {destroy();}

		bool create(int r, int c)
		{
			destroy();
			if(r <= 0 || c <= 0)
				return false;

			size = r*c;
			columns = c;
			rows = r;
			m = new float[size];
			return true;
		}
		bool resize(int r, int c)
		{
			if(!m)
				return false;

			float *temp;
			temp = new float[r*c];
			for(int x=0; x<c; ++x)
			{
				for(int y=0; y<r; ++y)
				{
					if(y < rows && x < columns)
						temp[y*c + x] = getM(y,x);
					else
						temp[y*c+x] = 0;
				}
			}

			delete[] m;
			m = temp;

			size = r*c;
			columns = c;
			rows = r;
			return true;
		}
		void destroy()
		{
			if(m)
				delete[] m;
			m = NULL;
			size = 0;
			columns = 0;
			rows = 0;
		}

		float get(int n)
		{
			if(!m)
				return 0;
			if(n < 0 || n >= size)
				return 0;

			return weight(m[n]);
		}
		float get(int r, int c)
		{
			if(!m)
				return 0;
			int index = r*columns + c;

			if(index < 0 || index >= size)
				return 0;

			return weight(m[index]);
		}

		float getDirect(int n)
		{
			if(!m)
				return 0;
			if(n < 0 || n >= size)
				return 0;

			return (m[n]);
		}
		float getDirect(int r, int c)
		{
			if(!m)
				return 0;
			int index = r*columns + c;

			if(index < 0 || index >= size)
				return 0;

			return (m[index]);
		}

		void set(int n, float value)
		{
			if(!m)
				return;
			if(n < 0 || n >= size)
				return;

			m[n] = value;
		}
		void set(int r, int c, float value)
		{
			if(!m)
				return;
			int index = r*columns + c;

			if(index < 0 || index >= size)
				return;

			if(fabs(value) < 0.0001f)
				value = 0;

			m[index] = value;
		}

		void zero()
		{
			if(!m)
				return;

			for(int i=0; i<size; ++i)
				m[i] = 0.0f;
		}

		void print(bool direct=false)
		{
			if(!m)
				return;
#ifdef MTENN_DEBUG
			std::cout << "U-Learn Matrix:\n";
			for(int r=0; r<rows; ++r)
			{
				std::cout << r << ":\t";
				for(int c=0; c<columns; ++c)
				{
					if(direct)
						std::cout << m[r*columns+c] << "\t";
					else
						std::cout << get(r,c) << "\t";
				}

				std::cout << "\n";
			}
			std::cout << "\n";
#endif
		}
	private:
		float getM(int r, int c)
		{
			if(!m)
				return 0;
			int index = r*columns + c;

			if(index < 0 || index >= size)
				return 0;

			return m[index];
		}
		float weight(float x)
		{
			//return x;
			const float M = 10;
			return (float)(2*M/(1+exp(-x/10.0)) - M);
		}
		float *m;
		int size, columns, rows;
	};

	class ULearn : public STM
	{
	public:
		ULearn()
		{
			actionSize = 0;
			stateSize = 0;
		}
		bool setup(int numInputs, int numOutputs)
		{
			return setupByPow2(powerOf2(numInputs), powerOf2(numOutputs));
		}
		bool setupByPow2(int _stateSize, int _actionSize)
		{
			stateSize = _stateSize;
			actionSize = _actionSize;
			if(!U_Matrix.create(stateSize, actionSize))
				return false;

			U_Matrix.zero();

			return true;
		}

		void addToMemory(const std::vector <float>& inputs, const std::vector <float>& outputs, float reward)
		{
#ifdef MTENN_STM_DEBUG
			std::cout << "Adding vector to memory\n";
#endif
			int state = convertVectorToIndex(inputs, 0.5f);
			int action = convertVectorToIndex(outputs, 0.5f);

			addR(state, action, reward);
		}

		bool resize(int _stateSize, int _actionSize)
		{
			if(!U_Matrix.resize(_stateSize, _actionSize))
				return false;

			stateSize = _stateSize;
			actionSize = _actionSize;

			return true;
		}

		void addR(int state, int action, float deltaR)
		{
			U_Matrix.set(state, action, U_Matrix.getDirect(state,action)+deltaR);
		}

		void copy(ULearn& U)
		{
			copy(&U);
		}
		void copy(ULearn* U)
		{
			if(!U)
				return;

			if(U->getActionSize() != getActionSize() || U->getStateSize() != getStateSize())
				return;

			for(int i=0; i<getStateSize(); ++i)
			{
				for(int j=0; j<getActionSize(); ++j)
				{
					setDirect(i, j, U->getDirect(i,j));
				}
			}
		}

		void add(ULearn& U)
		{
			if(U.getActionSize() != getActionSize() || U.getStateSize() != getStateSize())
				return;

			for(int i=0; i<getStateSize(); ++i)
			{
				for(int j=0; j<getActionSize(); ++j)
				{
					setDirect(i, j, getDirect(i,j) + U.getDirect(i,j));
				}
			}
		}

		void subtract(ULearn& U)
		{
			if(U.getActionSize() != getActionSize() || U.getStateSize() != getStateSize())
				return;

			for(int i=0; i<getStateSize(); ++i)
			{
				for(int j=0; j<getActionSize(); ++j)
				{
					setDirect(i, j, getDirect(i,j) - U.getDirect(i,j));
				}
			}
		}

		void clear()
		{
			for(int i=0; i<getStateSize(); ++i)
			{
				for(int j=0; j<getActionSize(); ++j)
				{
					setDirect(i, j, 0);
				}
			}
		}

		float getU(int state, float mixedAction)
		{
			float utility = 0.0f;
			for(int i=0; i<actionSize; ++i)
				utility = utility + (delta(mixedAction, (float)i)*U_Matrix.get(state, i));

			return utility;
		}

		float getU(int state, const std::vector <float>& mixedAction)
		{
			int m = mixedAction.size();

			float sum = 0, product;
			int pi_j, k;
			for(int i=0; i<actionSize; ++i)
			{
				product = U_Matrix.get(state, i);
				k = 1;
				for(int j=0; j<m; ++j)
				{
					pi_j = (i & k) / k;
					k = k << 1;
					if(pi_j == 1)
						product *= (mixedAction[j]);
					else
						product *= (1-mixedAction[j]);
				}

				sum += product;
			}

			return sum;
		}

		float getDirect(int r, int c)
		{
			return U_Matrix.getDirect(r,c);
		}
		float getDirect(int n)
		{
			return U_Matrix.getDirect(n);
		}
		void setDirect(int r, int c, float value)
		{
			U_Matrix.set(r, c, value);
		}

		float getRowSum(unsigned int row)
		{
			float sum = 0;
			for(int i=0; i<actionSize; ++i)
				sum += (float)fabs(U_Matrix.get(row, i));

			return sum;
		}

		float getMixedAction(const std::vector<float>& action)
		{
			float mixedAction = 0.0f;
			int power = 1;
			for(unsigned int i=0; i<action.size(); ++i)
			{
				mixedAction = mixedAction + (power * action[i]);
				power = power << 1;
			}

			return mixedAction;
		}

		int convertVectorToIndex(const std::vector <int>& v)
		{
			int power = 1;
			int result = 0;
			for(unsigned int i=0; i<v.size(); ++i)
			{
				result += (power*v[i]);
				power = power << 1;
			}

			return result;
		}
		int convertVectorToIndex(const std::vector <float>& v, float threshold)
		{
			int power = 1;
			int result = 0;
			int v_int=0;
			for(unsigned int i=0; i<v.size(); ++i)
			{
				if(v[i] >= threshold)
					v_int = 1;
				else
					v_int = 0;

				result += (power*v_int);
				power = power << 1;
			}

			return result;
		}

		void print()
		{
			U_Matrix.print();
		}

		int powerOf2(int x)
		{
			int i = 1;
			for(int y=0; y<x; ++y)
				i = i << 1;

			return i;
		}

		int getActionSize() {return actionSize;}
		int getStateSize() {return stateSize;}

		void multiplyScalar(float scalar)
		{
			for(int r=0; r<stateSize; ++r)
			{
				for(int c=0; c<actionSize; ++c)
					U_Matrix.set(r, c, U_Matrix.getDirect(r,c)*scalar);
			}
		}

		float fitness(ENN* network);
		float maxFitness();
		float minFitness();
		void degrade(float scalar) {multiplyScalar(scalar);};
	private:
		float delta(float x, float y)
		{
			float f = 1 - fabs(x-y);
			if(f > 0.0f)
				return f;
			return 0.0f;
		}
		float weightFct(float utility)
		{
			const float M = 5.0f;
			if(utility > M)
				return M;
			else if(utility < -M)
				return -M;
			return utility;
		}

		Matrix U_Matrix;
		int actionSize, stateSize;
	};
}

#endif