/*
timer.h
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef WOK_TIMER_H_
#define WOK_TIMER_H_

#include <time.h>

namespace wok
{
	class Timer
	{
	public:
		class Manager
		{
			friend class Timer;

			Manager() : pauseTime(0), paused(false), totalPauseTime(0) {};

			int getPauseTime() const
			{
				if(paused)
					return totalPauseTime + (clock()-pauseTime);

				return totalPauseTime;
			}
		public:
			static Manager& instance()
			{
				static Manager manager;
				return manager;
			}

			void pause(bool _pause)
			{
				if(_pause && !paused)
				{
					pauseTime = clock();
					paused = true;
				}
				else if(!_pause && paused)
				{
					totalPauseTime += clock() - pauseTime;
					paused = false;
				}
			}
		private:
			clock_t pauseTime;
			clock_t totalPauseTime;
			bool paused;
		};

		Timer() : msTime(0), stopTime(0), pauseTime(0) {};
		Timer(int _msTime) : msTime(_msTime), stopTime(0), pauseTime(0) {};

		void start()
		{
			stopTime = clock()+convertMSToTicks(msTime);
			pauseTime = Manager::instance().getPauseTime();
		};
		void start(int _msTime)
		{
			msTime = _msTime;
			start();
		};
		void set(int _msTime)
		{
			msTime = _msTime;
		};
		void stop()
		{
			stopTime = 0;
		};

		bool done() const
		{
			clock_t stopTime2 = stopTime + Manager::instance().getPauseTime() - pauseTime;
			return (stopTime2 <= clock());
		};

		float percentComplete()
		{
			if(msTime == 0)
				return 1;
			clock_t t = stopTime + Manager::instance().getPauseTime() - pauseTime - clock();
			float percent = (float) t;
			percent = percent / (float)(convertMSToTicks(msTime));

			percent = 1 - percent;

			if(percent < 0)
				percent = 0;
			if(percent > 1)
				percent = 1;

			return percent;
		}

		bool isRunning() const {return stopTime != 0;}

	private:
		clock_t convertMSToTicks(int _msTime)
		{
			return CLOCKS_PER_SEC*_msTime/1000;
		}
		clock_t stopTime, msTime, pauseTime;
	};
}

#endif