#ifndef SIMULATION_H_
#define SIMULATION_H_

/*
simulation.h Written by Walter O. Krawec
Demonstration of software described in the article

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

/*
To use an ENN, it is recommended to use ENNInterface in the following manner:

1) First create your instincts by inheriting MTENN::ENNInterface::Instinct
   a) One of which should be a "Boredom" instinct fashioned as below.
      In particular, it should produce a "correct" trigger if robot has been
	  moving around for a certain amount of time; an "incorrect" trigger if
	  the robot has been sitting still for too long. The "correct" trigger
	  is important for the DP system to function correctly.
2) Define your reflexes by inheriting from MTENN::ENNInterface::Reflex
3) Decide which STM to use (e.g. U-Learn or EL-Learn) and declare a memory object
   a) Example: MTENN::ULearn U; // if using U-Learning
   b)          MTENN::EvolutionList EL; // if using EL-Learning
4) Define an ENNInterface:
   a) Example: MTENN::ENNInterface::Interface ai;
5) Define instances of each instinct and reflex you use
6) Add reflexes to the correct instincts (via INSTINCT.addReflex(&REFLEX))
7) Add Instincts to Interface (via INTERFACE.addInstinct(&INSTINCT))
   a) See setupRobot() for examples
8) Run the ENN
   a) First setup using ai.setup(...) (see setupRobot())
   b) Repeat the following (see mainLoop()):
      i)   Use ai.setInput(INDEX, VALUE)
      ii)  ai.run()
      iii) ai.getOuput(INDEX)
      iv)  ai.mainLoop();	// call this to run DP manager and also to evolve on occasion.
*/

// Using SRS10 here, but you can use a different simulation if you like
#include "SRS10\SRS10Client.h"

#include "ENN\ENNInterface.h"
#include "ENN\ULearn.h"
#include "ENN\EL.h"

SRS10::Client simClient;	// main network interface to SRS10 simulator

// as per the directions, first define the instincts and reflexes::
class UserInput : public MTENN::ENNInterface::Instinct
{
public:
	UserInput(bool _correct) : correct(_correct), trigger(false) {};

	virtual float getDPScaleFactor() { return 1.0f/10.0f; };
	virtual float getReward() {return 2.0f;}

	virtual bool isCorrect() {return correct;}
	virtual bool isTriggered() {return trigger;};
	virtual void reset() {trigger = false;};

	virtual void printName() {std::cout << "user";};

	void set() {trigger = true;}
private:
	bool trigger;
	bool correct;
};

class Boredom : public MTENN::ENNInterface::Instinct
{
public:
	Boredom() : client(NULL), lastPosX(0), lastPosY(0), c(false) {checkTime.start(10000);};
	void setClient(SRS10::Client* c) {client = c;};
	void setRobotName(const std::string& name) {robotName = name;}

	virtual float getDPScaleFactor() {if(c == true)return 1.0f/1.5f;else return 1.0f/5.0f;};
	virtual float getReward() {if(c==true)return 2.0f;else return 4.0f;}
	virtual void printName() {std::cout << "boredom " << (int)c;};
	virtual bool isCorrect() {return c;}
	virtual bool isTriggered()
	{
		if(checkTime.done())
		{
			if(!client)
				return false;
			float distance = (lastPosX - client->getRobotXPos(robotName))*(lastPosX - client->getRobotXPos(robotName));
			distance += (lastPosY - client->getRobotYPos(robotName))*(lastPosY - client->getRobotYPos(robotName));

			lastPosX = client->getRobotXPos(robotName);
			lastPosY = client->getRobotYPos(robotName);

			if(distance <= 0.01f)
			{
				c = false;
				return true;
			}
			checkTime.start();
			c = true;
			return true;
		}

		return false;
	}

	virtual void reset() {checkTime.start();}

private:
	SRS10::Client* client;
	wok::Timer checkTime;
	float lastPosX, lastPosY;
	std::string robotName;

	bool c;
};

class Bump : public MTENN::ENNInterface::Instinct
{
public:
	Bump() : client(NULL), trigger(false) {checkTime.start(150);resetTime.start(5000);}
	void setClient(SRS10::Client* c) { client = c; };
	void setRobotName(const std::string& _robotName) {robotName = std::string(_robotName);}

	virtual float getDPScaleFactor() {return 1.0f/10.0f;};
	virtual float getReward() {return 5.0f;}
	virtual void printName() {std::cout << "bump";};
	virtual bool isCorrect() {return false;}
	virtual bool isTriggered()
	{
		return trigger;
		if(!checkTime.done())
			return false;
		if(!client)
			return false;

		if(client->getSensor(robotName, "collision") >= 0.5f)
			return true;

		checkTime.start();
		return false;
	}

	void set()
	{
		if(!resetTime.done()) return;
		trigger = true;
		resetTime.start();
	}

	virtual void reset() {trigger = false;}

private:
	wok::Timer checkTime, resetTime;
	SRS10::Client* client;
	std::string robotName;

	bool trigger;
};

class BackupReflex : public MTENN::ENNInterface::Reflex
{
public:
	BackupReflex() : client(NULL) {};
	BackupReflex(SRS10::Client* c, const std::string& _robotName) : client(c), robotName(_robotName) {};

	void setRobotName(const std::string& _robotName) {robotName = std::string(_robotName);};
	void setClient(SRS10::Client* c) {client = c;};

	virtual void trigger()
	{
		if(!client)
			return;

		client->moveCommand(robotName, 2);
		wok::Timer t;
		t.start(1000);
		while(!t.done());
		if(rand()%100 < 50)
			client->moveCommand(robotName, 3);
		else
			client->moveCommand(robotName, 4);
		t.start(1500+rand()%1000);
		while(!t.done());
		client->moveCommand(robotName, 0);
	}
private:
	SRS10::Client* client;
	std::string robotName;
};

Bump bump;
BackupReflex backup;
Boredom boredom;

MTENN::EvolutionList EL;
MTENN::ULearn U;
MTENN::ENNInterface::Interface ai;

wok::Timer timer;

UserInput userCorrect(true), userIncorrect(false);

// Call this to connect to SRS10 (run the simulator first of course)
bool connectToSRS10(char* IPAddress, int port=50010)
{
	return simClient.connectToHost(IPAddress, port);
}

// next call this to setup the robot; this will setup a robot with 3 IR sensors
void setupRobot()
{
	bool useEL = false;
	std::cout << "Use EL-Learning (if you type 'n', U-Learning will be used which is recommended)? (y/n) ";
	char c = std::cin.get();
	if(c == 'y')
		useEL = true;

	EL.setBinaryOutputMode(true);

	// setup the network with 3 inputs and 2 outputs, 0 hidden,
	// max DP size of 100, 10 iterations between DP points, and 125 iterations between calls to evolve()
	if(useEL)
		ai.setup(&EL, 3, 2, 0, 100, 10, 125);
	else
		ai.setup(&U, 3, 2, 0, 100, 10, 125);

	simClient.addRobot("robot1");	// a single client may control multiple robots, but each must have it's own name
									// multiple processes may each have a robot with the same name however
									// so you may run this program multiple times at once and there will be
									// no conflict despite the fact that each robot has the same name.
									// in other words: robot name must be unique only to a particular instance
									// of SRS10::Client

	// add three IR sensors; requires name of robot (setup using "addRobot")
	// a unique (to each robot, so different named robots may have the same named sensor) sensor name
	// the range (0.7 here), and the angle (in degrees relative to 0 being "forward")
	simClient.addIRSensor("robot1", "ir_left", 0.7f, 50);
	simClient.addIRSensor("robot1", "ir_center", 0.7f, 0);
	simClient.addIRSensor("robot1", "ir_right", 0.7f, -50);

	// setup the instincts
	boredom.setClient(&simClient);
	backup.setClient(&simClient);
	bump.setClient(&simClient);

	boredom.setRobotName("robot1");
	backup.setRobotName("robot1");
	bump.setRobotName("robot1");

	bump.addReflex(&backup);

	// add instincts to the ENN interface (so it's all automatically called when needed)
	ai.addInstinct(&userCorrect);
	ai.addInstinct(&userIncorrect);
	ai.addInstinct(&bump);
	ai.addInstinct(&boredom);

	// and set the timer to poll the simulator ever 150ms.
	timer.start(150);
}

// call this as often as possible; will only do something
// when timer is finished (default every 150ms)
void mainLoop()
{
	if(!timer.done())
		return;

	// Usually you should poll sensors first; this will grab the latest
	// data from the simulator regarding the sensors
	// needs the robot's name
	simClient.pollSensors("robot1");

	// use "getSensor(ROBOT_NAME, SENSOR_NAME) to read the value (from local data,
	// so if you didn't "pollSensors" first, you'll be using old data from the last poll)

	// SENSOR_NAME=="collision" is a special name used for track or other robot collisions
	int b = simClient.getSensor("robot1", "collision");
	if(b == 1)
		bump.set();	// we bumped
	
	// get left, right, center IR values
	int ir_left = simClient.getSensor("robot1", "ir_left");
	int ir_center = simClient.getSensor("robot1", "ir_center");
	int ir_right = simClient.getSensor("robot1", "ir_right");

	// feed these into the ENN
	ai.setInput(0, (float)ir_left);
	ai.setInput(1, (float)ir_center);
	ai.setInput(2, (float)ir_right);

	// run the ENN
	ai.runNetwork();

	// get ENN outputs
	int TL = ai.getOutputInt(0);
	int TR = ai.getOutputInt(1);

	// lastCommand is going to be, later on, used as a motor command
	// a command of "0" is STOP
	// 1 is FORWARD
	// 2 is BACK (not used here)
	// 3 is LEFT
	// 4 is RIGHT
	int lastCommand = 0;
	if(TL == 1 && TR == 0)
		lastCommand = 3;
	else if(TL == 0 && TR == 1)
		lastCommand = 4;
	else if(TL == 1 && TR == 1)
		lastCommand = 1;
	else
		lastCommand = 0;

	std::cout << "\r(" << ir_left << ", " << ir_center << ", " << ir_right << ") -> (" << TL << ", " << TR << ")          ";

	// Process Events
	// Events are sent by the simulator and they include user input (e.g. clapping) or picking up "food" items
	simClient.pollEvents("robot1");

	// After polling events, we may have a large list of things.  Each event specifies things like
	// type of event
	// where it happened (where was the robot)
	// what was the state (I/O values)

	// you need to iterate through them all before polling the events again else you'll lose the data.
	SRS10::Event e;
	std::vector <float> inputs, outputs;
	inputs.resize(ai.getNetwork().getNumInputs());
	outputs.resize(ai.getNetwork().getNumOutputs());

	// just keep calling "getNextEvent" until it returns false
	while(simClient.getNextEvent("robot1", e))
	{
		switch(e.lastMoveCommand)
		{
		case 1: outputs[0] = 1;
			outputs[1] = 1;
			break;
		case 3:
			outputs[0] = 1;
			outputs[1] = 0;
			break;
		case 4:
			outputs[0] = 0;
			outputs[1] = 1;
			break;
		default:
			outputs[0] = 0;
			outputs[1] = 0;
		}

		inputs[0] = (float)simClient.getEventSensorValue("ir_left", e);
		inputs[1] = (float)simClient.getEventSensorValue("ir_center", e);
		inputs[2] = (float)simClient.getEventSensorValue("ir_right", e);

		if(e.type == 0 || e.creator == 2)
		{
			int moveCommand = 0;
			switch(simClient.getEventDirection(e))
			{
			case SRS10::Client::Forward:
				outputs[0] = 1;
				outputs[1] = 1;
				moveCommand = 1;
				break;
			case SRS10::Client::Left:
				outputs[0] = 1;
				outputs[1] = 0;
				moveCommand = 3;
				break;
			case SRS10::Client::Right:
				outputs[0] = 0;
				outputs[1] = 1;
				moveCommand = 4;
				break;
			default:
				outputs[0] = 0;
				outputs[1] = 0;
				moveCommand = 0;
			}

			ai.getDP().addSpecificToSTM(inputs, outputs, 2.0f);
			wok::Timer delay;
			simClient.moveCommand("robot1", moveCommand);
			delay.start(1000);
			while(!delay.done());
		}
		else if(e.type == -1)
		{
			userIncorrect.set();
			ai.getDP().addSpecificToSTM(inputs, outputs, -2.0f);
		}
		else if(e.type == 1)
		{
			userCorrect.set();
			ai.getDP().addSpecificToSTM(inputs, outputs, 2.0f);
		}
		else if(e.type == 2)
		{
			userCorrect.set();
		}
	}

	// moveCommands are as described above (with "lastCommand" setting)

	simClient.moveCommand("robot1", 0);				// stop the robot before calling mainLoop incase an evolve() is called which takes time
	ai.mainLoop();
	simClient.moveCommand("robot1", lastCommand);	// now issue the movement command

	// and restart the timer for next iteration
	timer.start();
}

#endif