/*
Entity.h
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// Entities are small objects that may be "picked up" by a robot.  When a robot moves over an entity, an Event is triggered
// describing the type of entity ran over (positive or negative).

#ifndef SRS10_ENTITY_H_
#define SRS10_ENTITY_H_

#include "Sockets\Timer.h"

#ifdef WIN32
#include <Windows.h>
#endif

#include <gl/gl.h>		// Header File For The OpenGL32 Library
#include <gl/glu.h>		// Header File For The GLu32 Library

#ifndef WIN32
#include <GLUT/glut.h>		// Header File For The GLut Library
#endif

#include <vector>
#include <math.h>

namespace SRS10
{
	class Entity
	{
	public:
		Entity() : x(0), y(0), value(0), lifeTimeMS(-1), alive(true) {};
		
		Entity(float _x, float _y, int _value, int _lifeTime, int _respawnTime) : x(_x), y(_y), value(_value)
		{
			alive = true;
			lifeTimeMS = _lifeTime;
			lifeTime.start(lifeTimeMS);
			respawnTime.start(_respawnTime);
		}
		
		void draw()
		{
			if(!alive)
			{
				if(respawnTime.done())
				{
					alive = true;
					lifeTime.start();
				}
				else
					return;
			}
			
			if(lifeTime.done() && lifeTimeMS > 0)
			{
				alive = false;
				respawnTime.start();
				return;
			}
			
			const float size = 0.1f;
			
			if(value > 0)
				glColor4f(0,1,0,1);
			else if(value < 0)
				glColor4f(1,0,0,1);
			else
				glColor4f(1,1,1,1);
			
			glBegin(GL_QUADS);
				glVertex3f(x, y, 0);
				glVertex3f(x+size, y, 0);
				glVertex3f(x+size, y+size, 0);
				glVertex3f(x, y+size, 0);
			glEnd();
			
			glColor4f(1,1,1,1);
		}
		
		bool checkCollision(float _x, float _y, float r)
		{
			if(!alive)
				return false;
			
			const float d = 0.1f+r;
			if(sqrtf( (_x-x)*(_x-x) + (_y-y)*(_y-y) ) <= d)
			{
				alive = false;
				respawnTime.start();
				return true;
			}
			return false;
		}
		
		int getValue() {return value;};
		
	//private:
		float x, y;
		int value;
		bool alive;
		
		int lifeTimeMS;
		wok::Timer lifeTime, respawnTime;
	};
	
	class EntityManager
	{
	public:
		void add(Entity e)
		{
			entityList.push_back(e);
		}
		
		void draw()
		{
			for(int i=0; i<entityList.size(); ++i)
				entityList[i].draw();
		}
		
		bool checkCollision(Entity& e, float x, float y, float r, bool clearIfCollide=true)
		{
			if(entityList.empty())
				return false;
			
			for(int i=0; i<entityList.size(); ++i)
			{
				if(entityList[i].checkCollision(x,y,r))
				{
					e = entityList[i];
					return true;
				}
			}

			return false;
		}
		
		static EntityManager& instance()
		{
			static EntityManager manager;
			return manager;
		}

		std::vector <Entity>& getEntityList() {return entityList;};
	private:
		std::vector <Entity> entityList;
	};
}
#endif