/*
Math3D.h
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef MATH3D_H_
#define MATH3D_H_

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

namespace math3D
{

	class CVector2D
	{
	public:
		CVector2D(){x=0; y=0;};
		CVector2D(float _x, float _y){x = _x; y = _y;};
		void Construct(float rot, float size);	// constructs a 2d vector given its rotation and size (z = 0)
		void rotate(float Rotation)
		{
			// rotates a point around origin with x and y being the distance from origin
			float rot = -Rotation;
			float x2 = 0;
			float y2 = 0;

			if(x < 0)
			{
				rot += 90.0f;
			}
			else
			{
				x = x * -1.0f;
				rot -= 90.0f;
			}
			x2 = sinf(rot*0.0174532925f) * (x);
			y2 = cosf(rot*0.0174532925f) * (x);

			rot = -Rotation;
			if(y < 0)
			{
				rot -= 180.0f;
				y = y * -1.0f;
			}
			x2 += sinf(rot*0.0174532925f) * (y);
			y2 += cosf(rot*0.0174532925f) * (y);

			x = x2;
			y = y2;
		}

		CVector2D operator+ (const CVector2D& a) const
		{
			return CVector2D(x+a.x, y+a.y);
		}

		float& operator[] (unsigned i){return (i==0 ? x : y);};

		float x,y;
	};

	class CVector3D
	{
	public:
		CVector3D(){x=0; y=0; z=0;};
		CVector3D(float _x, float _y, float _z){x = _x; y = _y; z = _z;};
		CVector3D(const CVector3D& v){x = v.x; y = v.y; z = v.z;};
		CVector3D(const CVector2D& v) : x(v.x), y(v.y), z(0.0f) {};
		float& operator[] (unsigned i){return (i==0 ? x : i==1 ? y : z);};

		unsigned int GetTile(float tileSize, unsigned int xSize)
		{
			float fx = (float)x / tileSize;
			float fy = (float)y / tileSize;
			unsigned int tile = (unsigned int)fy*xSize;
			tile += (unsigned int) fx;
			return tile;
		};

		void GetTilePosition(unsigned int tile, unsigned int xSize, float tileSize)
		{
			x = (tile%xSize)*tileSize;
			y = (tile/xSize)*tileSize;
			z = 0;
		};

		CVector3D operator+ (const CVector3D& a) const
		{
			return CVector3D(x+a.x, y+a.y, z+a.z);
		}

		float x,y,z;
	};

	class CMath3D
	{
	public:
		static float FindVectorAngle(CVector2D pos);
		static float FindDistanceSq(const CVector2D& v1, const CVector2D& v2);
		static float FindDistanceSq(const CVector3D& v1, const CVector3D& v2);
		static float FindDistanceSq2D(const CVector3D& v1, const CVector3D& v2);
		static float FindDistanceQuick(const CVector3D& v1, const CVector3D& v2) {return FindDistanceSq(v1, v2);};

		static CVector2D FindRandomPos(const CVector2D& center, float radius);
		static CVector3D FindRandomPos(const CVector3D& center, float radius);
		static CVector3D FindRandomPos3D(CVector2D& center, float radius);
		static float Rad2Deg(float rad);
		static float Deg2Rad(float deg);

		static float randFloat(float max)
		{
			return (rand()%100) * (max/100);
		};

		static float normalizeRotation(float rot)
		{
			int r = (int)rot;
			r %= 360;
			if(r < 0)
				r += 360;
			return (float)r;
		};

		// finds the distance (in world space) between two tiles (x1,y1) and (x2,y2):
		static float FindTileDistanceSq(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2, float TileSize);

		static int FindTileDistance(int tile1, int tile2, int xSize);
		static int FindRandomTile(int center, int radius, unsigned int sizeX, unsigned int sizeY);

		static int GetTileNumber(const CVector3D& pos, float tileSize, unsigned int xSize)
		{
			float fx = (float)pos.x / tileSize;
			float fy = (float)pos.y / tileSize;
			unsigned int tile = (unsigned int)fy*xSize;
			tile += (unsigned int) fx;
			return tile;
		};
		static int GetTileNumber(const CVector2D& pos, float tileSize, unsigned int xSize)
		{
			float fx = (float)pos.x / tileSize;
			float fy = (float)pos.y / tileSize;
			unsigned int tile = (unsigned int)fy*xSize;
			tile += (unsigned int) fx;
			return tile;
		};
		static math3D::CVector3D GetTilePosition(unsigned int tile, unsigned int xSize, float tileSize)
		{
			return math3D::CVector3D((tile%xSize)*tileSize, (tile/xSize)*tileSize, 0);
		};

	};

	class CBoundingSphere
	{
	public:
		CBoundingSphere() {};
		CBoundingSphere(float size) : m_Size(size) {};

		float get() const {return m_Size;};
		static bool collide(const CBoundingSphere& sphere1, const math3D::CVector3D& pos1, const CBoundingSphere& sphere2, const math3D::CVector3D& pos2);

	private:
		float m_Size;
	};

	inline
	void CVector2D::Construct(float rot, float size)
	{
		x = cosf(rot)*size;
		y = sinf(rot)*size;
	}

	inline
		CVector3D CMath3D::FindRandomPos3D(CVector2D& center, float radius)
	{
		math3D::CVector2D v = FindRandomPos(center, radius);
		return math3D::CVector3D(v.x, v.y, 0);
	}

	inline
	float CMath3D::FindDistanceSq(const CVector2D& v1, const CVector2D& v2)
	{
		return ((v2.x-v1.x)*(v2.x-v1.x)) + ((v2.y-v1.y)*(v2.y-v1.y));//(powf((v2.x-v1.x), 2) + powf((v2.y-v1.y), 2));
	}

	inline
	float CMath3D::FindDistanceSq(const CVector3D& v1, const CVector3D& v2)
	{
		return ((v2.x-v1.x)*(v2.x-v1.x)) + ((v2.y-v1.y)*(v2.y-v1.y)) + ((v2.z-v1.z)*(v2.z-v1.z));//(powf((v2.x-v1.x), 2) + powf((v2.y-v1.y), 2));
	}

	inline
	float CMath3D::FindDistanceSq2D(const CVector3D& v1, const CVector3D& v2)
	{
		return ((v2.x-v1.x)*(v2.x-v1.x)) + ((v2.y-v1.y)*(v2.y-v1.y));//(powf((v2.x-v1.x), 2) + powf((v2.y-v1.y), 2));
	}


	inline
	int CMath3D::FindTileDistance(int tile1, int tile2, int xSize)
	{
		int x=tile1/xSize;
		int y=tile1%xSize;

		int x2=tile2/xSize;
		int y2=tile2%xSize;

		x = abs(x-x2);
		y = abs(y-y2);

		if(y > x)
			x = y;

		return x;
	}

	inline
	float CMath3D::Rad2Deg(float rad)
	{
		return (rad*57.29577951308f);
	}

	inline
	float CMath3D::Deg2Rad(float deg)
	{
		return (deg*0.0174532925f);
	}

	inline
	int CMath3D::FindRandomTile(int center, int radius, unsigned int sizeX, unsigned int sizeY)
	{
		int centerX = center%sizeX, centerY = center/sizeX;
		int x = rand()%(radius*2)-radius+centerX;
		int y = rand()%(radius*2)-radius+centerY;

		int tile = y*sizeX+x;
		if(tile < 0 || tile >= (int)(sizeX*sizeY) || FindTileDistance(tile, center, sizeX) > radius)
			return -1;

		return tile;
	}



	inline
	bool CBoundingSphere::collide(const CBoundingSphere& sphere1, const math3D::CVector3D& pos1, const CBoundingSphere& sphere2, const math3D::CVector3D& pos2)
	{
		return false;
	}
}

#endif