/*
Shapes.h
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// graphical code - draws shapes used to construct robots, tracks, etc.

#ifndef SRS10_SHAPES_
#define SRS10_SHAPES_

#include <Windows.h>
#include "math3d.h"
#include <gl/gl.h>		// Header File For The OpenGL32 Library
#include <gl/glu.h>		// Header File For The GLu32 Library
//#include <GLUT/glut.h>		// Header File For The GLut Library

#include <string>
#include <fstream>
#include <vector>

#include <list>

#include "TextParser.h"
#include "Entity.h"

namespace SRS10{
namespace graphics{

	struct CollisionCircle
	{
		float x,y,size;
	};
	
	struct Message
	{
		Message() : useDefaultPos(true), x(0), y(0) {};
		std::string message;
		time_t endTime;
		
		float x, y;
		bool useDefaultPos;
	};
	
	class GraphicsEngine
	{
	public:
		GraphicsEngine() : FPS(1), counter(1), simulationSpeed(0.5f) {newTime = time(NULL)+1;};
		static GraphicsEngine& instance()
		{
			static GraphicsEngine engine;
			return engine;
		};
		
		void printString(const std::string& str, float x, float y, float z, bool smallText=false)
		{
			glRasterPos3d(x,y,z);
			
			for(int i=0; i<str.size(); ++i)
			{
				//if(smallText)
			//		glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_10, str[i]);
			//	else
			//		glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24, str[i]);
			}
		}
		
		void updateFPS()
		{
			++counter;
			if(time(NULL) >= newTime)
			{
				FPS = counter*1;
				//FPS = 700;
				newTime = time(NULL)+1;
				
				counter = 0;
			}
		}
		
		void addMessage(const std::string& str, float x=0, float y=0, bool useDefaultPos=true, time_t seconds = 5)
		{
			Message msg;
			msg.message = str;
			msg.endTime = time(NULL) + seconds;
			msg.x = x;
			msg.y = y;
			msg.useDefaultPos = useDefaultPos;
			messages.push_back(msg);
		}
		
		void displayMessages(float x, float y, int max=10)
		{
			if(messages.empty())
				return;
			
			std::list <Message>::iterator Iter;
			for(Iter = messages.begin(); Iter != messages.end(); ++Iter)
			{
				if(Iter->useDefaultPos)
				{
				printString(Iter->message, x, y, -5);
				y -= .175f;
				
				-- max;
				
				if(max <= 0)
					break;
				
				if(time(NULL) >= Iter->endTime)
				{
					messages.erase(Iter);
					break;
				}
				}
			}
		}
		
		void displayFloatingMessages()
		{
			if(messages.empty())
				return;
			
			std::list <Message>::iterator Iter;
			for(Iter = messages.begin(); Iter != messages.end(); ++Iter)
			{
				if(!Iter->useDefaultPos)
				{
					printString(Iter->message, Iter->x, Iter->y, 0, false);
					
					if(time(NULL) >= Iter->endTime)
					{
						messages.erase(Iter);
						break;
					}
				}
			}
			
		}
		
		int getFPS() { return FPS; };

		float getSimulationSpeed() {return simulationSpeed;};
		void setSimulationSpeed(float value) {simulationSpeed = value;};
		void changeSimulationSpeed(float deltaSpeed) {simulationSpeed += deltaSpeed;};
		
	private:
		time_t newTime;
		int FPS, counter;

		float simulationSpeed;
		
		std::list <Message> messages;
	};
	
	class Shape
	{
	public:
		Shape() : solidObject(true) {};
		virtual bool collision(const CollisionCircle& circle){return false;}
		virtual void draw(float offsetX, float offsetY, float rot)=0;

		bool solidObject;
	};

	class Line : public Shape
	{
	public:
		Line() : x1(0), y1(0), x2(0), y2(0) {};
		
		virtual void draw(float offsetX, float offsetY, float rot)
		{
			glBegin(GL_LINES);
			glVertex3f(x1+offsetX, y1+offsetY, 0);
			glVertex3f(x2+offsetX, y2+offsetY, 0);
			glEnd();
		}
		
		virtual bool collision(const CollisionCircle& circle)
		{
			float dist = sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1));
			int steps = dist/circle.size + 1;
			float dx = (x2-x1)/steps, dy = (y2-y1)/steps;
			float x = x1, y = y1;
			for(int i=0; i<steps; ++i)
			{
				float d = sqrt( (x-circle.x)*(x-circle.x) + (y-circle.y)*(y-circle.y) );
				if(d <= circle.size)
					return true;
				
				x += dx;
				y += dy;
			}
			return false;
			/*float dist = (fabs((x2-x1)*(y1-circle.y) - (x1-circle.x)*(y2-y1))/sqrt( (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) ));
			if(dist <= circle.size)
				return true;
			return false;
			
			float xp1 = x1-circle.x;
			float xp2 = x2-circle.x;
			float yp1 = y1-circle.y;
			float yp2 = y2-circle.y;
			
			float dx = xp2-xp1;
			float dy = yp2-yp1;
			float dr = sqrtf(dx*dx+dy*dy);
			float D = xp1*yp2 - xp2*yp1;
			float d = circle.size*circle.size*dr*dr - D*D;
			if(d >= 0)
				return true;
			return false;*/
			
		}
		
		float x1,y1, x2,y2;
	};

	class BoldLine : public Shape
	{
	public:
		BoldLine() : x1(0), y1(0), x2(0), y2(0) {};
		
		virtual void draw(float offsetX, float offsetY, float rot)
		{
			glColor4f(0,1,0,1);
			glBegin(GL_LINES);
				glVertex3f(x1+offsetX, y1+offsetY, 0);
				glVertex3f(x2+offsetX, y2+offsetY, 0);
			glEnd();
			glColor4f(1,1,1,1);
		}
		
		virtual bool collision(const CollisionCircle& circle)
		{
			float dist = sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1));
			int steps = dist/circle.size + 1;
			float dx = (x2-x1)/steps, dy = (y2-y1)/steps;
			float x = x1, y = y1;
			for(int i=0; i<steps; ++i)
			{
				float d = sqrt( (x-circle.x)*(x-circle.x) + (y-circle.y)*(y-circle.y) );
				if(d <= circle.size)
					return true;
				
				x += dx;
				y += dy;
			}
			return false;
		}
		
		float x1,y1, x2,y2;
	};
	
	class Circle : public Shape
	{
	public:
		Circle() : x1(0), y1(0), radius(1){};
		
		virtual void draw(float offsetX, float offsetY, float rot)
		{
			float x = offsetX, y = offsetY;
			int step = (int)(10.0f*radius);
			if(step < 10)
				step = 10;
			
			x += x1;
			y += y1;
			const float d = 6.2832f/step;
			math3D::CVector2D lastPoint,v2;
			lastPoint.Construct(0, radius);
			for(int i=1; i<step; ++i)
			{
				v2.Construct(i*d, radius);
				glBegin(GL_LINES);
				glVertex3f(v2.x+x, v2.y+y, 0);
				glVertex3f(lastPoint.x+x, lastPoint.y+y, 0);
				glEnd();
				
				lastPoint = v2;
			}
			
			v2.Construct(0, radius);
			glBegin(GL_LINES);
			glVertex3f(v2.x+x, v2.y+y, 0);
			glVertex3f(lastPoint.x+x, lastPoint.y+y, 0);
			glEnd();
		}
		
		float x1, y1, radius;
	};
	
	
	class Triangle : public Shape
		{
		public:
			virtual void draw(float x, float y, float rot)
			{
				math3D::CVector2D a,b,c;
				
				a.Construct(math3D::CMath3D::Deg2Rad(90+rot), baseSize/2.0f);
				a.x += x1+x;
				a.y += y1+y;
				
				b.Construct(math3D::CMath3D::Deg2Rad(-90+rot), baseSize/2.0f);
				b.x += x1+x;
				b.y += y1+y;
				
				c.Construct(math3D::CMath3D::Deg2Rad(0+rot), height);
				c.x += x1+x;
				c.y += y1+y;
				
				glBegin(GL_LINES);
				glVertex3f(a.x, a.y, 0); glVertex3f(b.x, b.y, 0);
				glVertex3f(b.x, b.y, 0); glVertex3f(c.x, c.y, 0);
				glVertex3f(c.x, c.y, 0); glVertex3f(a.x, a.y, 0);
				glEnd();
			}
			
			virtual bool collision(const CollisionCircle& circle)
			{
				return false;
			}
			
			float x1,y1,baseSize,height;
		};
	
	class ShapeCollection
		{
		public:
			~ShapeCollection() {cleanup();}
			void cleanup()
			{
				for(size_t i=0; i<shapeList.size(); ++i)
				{
					if(shapeList[i])
						delete shapeList[i];
				}
				
				shapeList.clear();
			}
			
			void draw(float x=0, float y=0, float rot=0)
			{
				for(size_t i=0; i<shapeList.size(); ++i)
				{
					if(shapeList[i])
						shapeList[i]->draw(x,y,rot);
				}
			}
			
			bool checkCollision(const CollisionCircle& circle)
			{
				for(size_t i=0; i<shapeList.size(); ++i)
				{
					if(shapeList[i] && shapeList[i]->solidObject && shapeList[i]->collision(circle))
						return true;
				}
				
				return false;
			}

			bool checkNonSolidCollision(const CollisionCircle& circle)
			{
				for(size_t i=0; i<shapeList.size(); ++i)
				{
					if(shapeList[i] && !shapeList[i]->solidObject && shapeList[i]->collision(circle))
						return true;
				}
				
				return false;
			}
			
			void add(Shape* newShape)
			{
				shapeList.push_back(newShape);
			}
			
			bool loadFromDataFile(const std::string& filename);
		private:
			std::vector <Shape*> shapeList;
		};	

}}
#endif