/*
Sockets.cpp
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// A really easy to use socket interface; works on Windows and Mac (probably others though only Windows and Mac were ever tested)
// Manages all the busy work of creating a client/server TCP connection

#include "Sockets.h"

#include <iostream>

namespace wok
{
namespace sockets
{
	void Packet::updateHeader()
	{
		if(bufferIndex < 5)
			return;
//		packetData[1] = (char)bufferIndex;
		
		uint32 dataToSend = htonl(bufferIndex);
		
		for(int i=0; i<sizeof(int); ++i)
		{
			packetData[i+1] = ((char*)&dataToSend)[i];
		}
	}
	bool Packet::packChar(char data)
	{
		if(bufferIndex >= MAX_PACKET_SIZE)
			return false;
		packetData[bufferIndex] = data;
		++bufferIndex;
		return true;
	}
	bool Packet::packInt(int data)
	{
		if(bufferIndex+sizeof(int) >= MAX_PACKET_SIZE)
			return false;

		uint32 dataToSend = htonl(data);
		
		for(int i=0; i<sizeof(int); ++i)
		{
			packetData[bufferIndex] = ((char*)&dataToSend)[i];
			++bufferIndex;
		}

		return true;
	}
	bool Packet::packString(const std::string& str)
	{
		if(str.size() >= 256)
			return false;
		if(bufferIndex+str.size()+1 >= MAX_PACKET_SIZE)
			return false;
		
		for(int i=1; i<=(int)str.size(); ++i)
			packChar(str[str.size()-i]);
		
		packChar((int)str.size());
		return true;
	}
	bool Packet::packFloat(float data)
	{
		std::stringstream strValue;
		strValue << data;
		return packString(strValue.str());
		/*if(bufferIndex+sizeof(float) >= MAX_PACKET_SIZE)
			return false;
		
		for(int i=0; i<sizeof(float); ++i)
		{
			packetData[bufferIndex] = ((char*)&data)[i];
			++bufferIndex;
		}*/

		return true;
	}
	char Packet::unpackChar()
	{
		if(bufferIndex < 1)
			return 0;
		
		--bufferIndex;
		char c = packetData[bufferIndex];
		return c;
	}
	int Packet::unpackInt()
	{
		if(bufferIndex < sizeof(int))
		   return 0;
		
		uint32 data=0;
		for(int i=1; i<=sizeof(int); ++i)
		{
			--bufferIndex;
			((char*)&data)[sizeof(int)-i] = packetData[bufferIndex];
		}
		
		return (int)ntohl(data);
	}
	void Packet::unpackString(std::string& str)
	{
		char buffer[257];
		int size = unpackChar();
		if(bufferIndex < size || size >= 256)
			return;
		
		for(int i=0; i<size; ++i)
		{
			buffer[i] = unpackChar();
			buffer[i+1] = 0;
		}
		
		str = std::string(buffer);
	}
	float Packet::unpackFloat()
	{
		/*if(bufferIndex < sizeof(float))
			return 0;
		
		float data=0;
		for(int i=1; i<=sizeof(float); ++i)
		{
			--bufferIndex;
			((char*)&data)[sizeof(float)-i] = packetData[bufferIndex];
		}
		
		return data;*/
		
		std::string strValue;
		unpackString(strValue);
		return (float)atof(strValue.c_str());
	}
	
	void Packet::debugPrintPacket()
	{
		for(int i=0; i<bufferIndex; ++i)
			std::cout << (int)packetData[i] << ", ";
	};
	
	bool Socket::ping()
	{
//		std::cout << "debug ping\n";
		if(socketID == -1)
			return false;
		
		Packet p;
		p.createHeader(0);
		p.packChar(0);
		p.updateHeader();
		if(!sendPacket(p))
			return false;
//		if(send(socketID, &c, 1, 0) != 1)
//			return false;
		
		return true;
	}
	
	bool Socket::sendPacket(Packet& p)
	{
		if(socketID == -1)
			return false;
		
		p.updateHeader();
		if(send(socketID, p.getPacketData(), p.packetSize(), 0) != p.packetSize())
			return false;
		
		return true;
	}
	void Socket::checkForData()
	{
		if(bufferIndex >= MAX_RECV_BUFFER || socketID == -1)
			return;
#ifdef WIN32
		timeval timeout;
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
		fd_set fd;
		FD_ZERO(&fd);
		FD_SET(socketID, &fd);
		if(select(0, &fd, NULL, NULL, &timeout) <= 0)
			return;
#endif
		
		int data = recv(socketID, (char*)(&recvBuffer)+bufferIndex, MAX_RECV_BUFFER-bufferIndex,0);
		if(data >= 1)
			bufferIndex += data;
//		if(data == -1 && (errno != EAGAIN && errno != EWOULDBLOCK))
//			std::cout << "debug = " << (int)errno << ".\n";
	}

	int Socket::nextPacketType()
	{
		if(bufferIndex == 0)
			return -1;

		return recvBuffer[0];
	}
	
	int Socket::grabNextPacket(Packet& packet)
	{
		if(bufferIndex < HEADER_SIZE)
			return -1;	// nothing here yet
		
		int packetType = recvBuffer[0];
//		int packetSize = recvBuffer[1];
		uint32 tempSize=0;
		for(int i=0; i<sizeof(int); ++i)
		{
			((char*)&tempSize)[i] = recvBuffer[i+1];
		}
		
		uint32 packetSize = ntohl(tempSize);
		if(bufferIndex < (int)packetSize)
			return -1;	// packet not complete yet
		
//		std::cout << "debug index = " << bufferIndex << "\n";
		packet.clear();	// make sure this packet is empty
		for(int i=0; i<(int)packetSize; ++i)
		{
			packet.packChar(recvBuffer[i]);
		}
		
		for(int i=0; i<(int)(MAX_RECV_BUFFER-packetSize); ++i)
			recvBuffer[i] = recvBuffer[i+packetSize];
		
		bufferIndex -= packetSize;

		if(packetType == 0)
		{
			processPacket(packet);
			packet.clear();
			return -1;
		}
		
//		std::cout << "debug index = " << bufferIndex << "\n";
		return packetType;
	}
	bool Socket::initWinSock()
	{
#ifdef WIN32
		WSADATA WSVersion;

		int error = WSAStartup(0x0202, &WSVersion);

		if(error)
			return false;

		if(WSVersion.wVersion != 0x0202)
		{
			WSACleanup();
			return false;
		}
#endif

		return true;
	}

	void Socket::processPacket(Packet& p)
	{
		int type = p.unpackChar();
		if(type == 1)
		{
			MultiPacket mp;
			mp.ID.set(p.unpackInt());
			mp.sendPosition = p.unpackInt();
			mp.type = p.unpackInt();
			mp.complete = false;
			multiPacketInbox.push_back(mp);
		}
		else if(type == 2)
		{
			wok::UID ID;
			ID.set(p.unpackInt());

			if(multiPacketOutbox.empty())
				return;
			std::list <MultiPacket>::iterator Iter;
			for(Iter = multiPacketOutbox.begin(); Iter != multiPacketOutbox.end(); ++Iter)
			{
				if(Iter->ID == ID)
				{
					sendMultiPacket(Iter);
					break;
				}
			}
		}
		else if(type == 3)
		{
			wok::UID ID;
			ID.set(p.unpackInt());

			if(multiPacketInbox.empty())
				return;
			std::list <MultiPacket>::iterator Iter;
			for(Iter = multiPacketInbox.begin(); Iter != multiPacketInbox.end(); ++Iter)
			{
				if(Iter->ID == ID)
				{
					//std::cout << "debug5\n";
					Packet temp;
					for(int i=HEADER_SIZE; i<p.packetSize(); ++i)
						temp.setBufferElement(i-HEADER_SIZE, p.getBufferElement(i));
					temp.setBufferIndex(p.packetSize()-HEADER_SIZE);

					Iter->packetList.push_back(temp);
					--Iter->sendPosition;
					if(Iter->sendPosition == 0)
					{
						Iter->complete = true;
						++numCompleteMP;
					}
					else
					{
						// request next section
						temp.clear();
						temp.createHeader(0);
						temp.packInt(ID.get());
						temp.packChar(2);
						sendPacket(temp);
					}

					break;
				}
			}
		}
	}

	void Socket::sendMultiPacket(std::list <MultiPacket>::iterator& Iter)
	{
		Packet packetToSend, temp;
		Iter->getPacket(Iter->sendPosition, temp);
		++Iter->sendPosition;
		packetToSend.createHeader(0);
		for(int i=0; i<temp.packetSize(); ++i)
			packetToSend.packChar(temp.getBufferElement(i));
		packetToSend.packInt(Iter->ID.get());
		packetToSend.packChar(3);
		sendPacket(packetToSend);

		if(Iter->sendPosition == Iter->packetList.size())
			multiPacketOutbox.erase(Iter);
	}

	void Socket::sendMultiPacket(MultiPacket& mp)
	{
		mp.sendPosition = 0;
		multiPacketOutbox.push_back(mp);

		Packet p;
		p.createHeader(0);
		p.packInt(mp.type);
		p.packInt((int)mp.packetList.size());
		p.packInt(mp.ID.get());
		p.packChar(1);
		sendPacket(p);
	}

	bool Socket::recvMultiPacket(MultiPacket& mp)
	{
		if(numCompleteMP <= 0 || multiPacketInbox.empty())
			return false;

		std::list <MultiPacket>::iterator Iter;
		for(Iter = multiPacketInbox.begin(); Iter != multiPacketInbox.end(); ++Iter)
		{
			if(Iter->complete)
			{
				mp = *Iter;
				--numCompleteMP;
				return true;
			}
		}

		return false;
	}

	void Socket::mainLoop()
	{
		if(socketID == -1)
			return;

		if(nextPacketType() == 0)
		{
			Packet p;
			grabNextPacket(p);
		}

		if(!multiPacketOutbox.empty())
		{
			std::list <MultiPacket>::iterator Iter = multiPacketOutbox.begin();
			if(Iter->sendPosition == 0)
			{
				sendMultiPacket(Iter);
			}
		}
	}

	bool Socket::shutdownWinSock()
	{
#ifdef WIN32
		WSACleanup();
#endif
		return true;
	}
	
	bool Host::setupHost(unsigned short portNumber, int _timeToPing)
	{
#ifdef WIN32
		const int MAXHOSTNAMELEN = 100;
#endif
		timeToPing = _timeToPing;
		shutdownHost();
		
		if(!mainSocket.setupSocket())
			return false;
		mainSocket.setNonBlocking();
		
		sockaddr_in socketAddr;
		hostent *hp;
		char name[MAXHOSTNAMELEN+1];
		
		memset(&socketAddr, 0, sizeof(struct sockaddr_in));
		gethostname(name, MAXHOSTNAMELEN);
		hp = gethostbyname(name);
		if(hp == NULL)
		{
			shutdownHost();
			return false;
		}
		
		socketAddr.sin_family = hp->h_addrtype;
		socketAddr.sin_port = htons(portNumber);
		if(bind(mainSocket.getSocketID(), (sockaddr*)&socketAddr, sizeof(struct sockaddr_in)) < 0)
		{
			shutdownHost();
			return false;
		}
		
		listen(mainSocket.getSocketID(), 5);
		
		pingClients.start(timeToPing);
		return true;
	}
	
	bool Host::lookForConnection()
	{
		if(mainSocket.getSocketID() == -1)
			return false;

		#ifndef WIN32
		int t;
#else
		SOCKET t;
		timeval timeout;
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
		fd_set fd;
		FD_ZERO(&fd);
		FD_SET(mainSocket.getSocketID(), &fd);
		if(select(0, &fd, NULL, NULL, &timeout) <= 0)
			return false;	// nothing to connect to

#endif

		t = accept(mainSocket.getSocketID(), NULL, NULL);
		
		if(t >= 0)
		{
			for(int i=0; i<MAX_CLIENTS; ++i)
			{
				if(clientSockets[i].getSocketID() == -1)
				{
					clientSockets[i].setSocketID(t);
					clientSockets[i].setNonBlocking();
					if(printLogToScreen)
						std::cout << "Client connected, ID = " << i << "\n";
					return true;
				}
			}
#ifndef WIN32
			close(t);
#else
			closesocket(t);
#endif
			return false;
		}
		return false;
	}
	
	void Host::mainLoop()
	{
		lookForConnection();
		mainSocket.checkForData();
		for(int i=0; i<MAX_CLIENTS; ++i)
			clientSockets[i].checkForData();
	
		for(int i=0; i<MAX_CLIENTS; ++i)
		{
			clientSockets[i].mainLoop();
			if(pingClients.done())
			{
				if(clientSockets[i].getSocketID() != -1 && !clientSockets[i].ping())
				{
					if(printLogToScreen)
						std::cout << "Client disconnected, ID = " << i << "\n";
					clientSockets[i].shutdownSocket();
				}
			}
		}
		if(pingClients.done())
			pingClients.start(timeToPing);
	}
	
	void Host::disconnectClient(int clientID)
	{
		if(clientID < 0 || clientID >= MAX_CLIENTS)
			return;
		
		clientSockets[clientID].shutdownSocket();
	}
	
	bool Host::dataWaiting(int clientID)
	{
		if(clientID < 0 || clientID >= MAX_CLIENTS)
			return false;
		return clientSockets[clientID].dataWaiting();
	}
	
	Socket* Host::getClientSocket(int clientID)
	{
		if(clientID < 0 || clientID >= MAX_CLIENTS)
			return NULL;
		
		return &clientSockets[clientID];
	}
	
	bool Host::sendPacket(Packet& p, int clientID)
	{
		if(clientID == -1)
		{
			for(int i=0; i<MAX_CLIENTS; ++i)
				clientSockets[i].sendPacket(p);
			
			return true;
		}
		else if(clientID >= 0 && clientID < MAX_CLIENTS)
			return clientSockets[clientID].sendPacket(p);
		
		return false;
	}
	
	bool Client::connectToHost(const char* IPAddress, unsigned short portNumber)
	{
		if(!mainSocket.setupSocket())
			return false;
		
		struct sockaddr_in sockAddr;
		
		memset(&sockAddr, 0, sizeof(struct sockaddr_in));
		
		sockAddr.sin_family = AF_INET;
		sockAddr.sin_port = htons(portNumber);
#ifndef WIN32
		inet_pton(AF_INET, IPAddress, &sockAddr.sin_addr);
#else
		sockAddr.sin_addr.s_addr = inet_addr(IPAddress);
#endif
		
		if (connect(mainSocket.getSocketID(), (struct sockaddr *)&sockAddr, sizeof(sockAddr)) == -1)
		{
			mainSocket.shutdownSocket();
			return false;
		}
		
		mainSocket.setNonBlocking();
		return true;
	}
	
	void Client::mainLoop()
	{
		mainSocket.checkForData();
		mainSocket.mainLoop();
	}
}
}