/*
Track.h
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// Manages the simulation track/environment

#ifndef SRS10_TRACK_
#define SRS10_TRACK_

#include "Shapes.h"
#include "Entity.h"
#include <vector>
#include "TextParser.h"
#include "Sockets\Timer.h"

namespace SRS10{
namespace track
{	
	class VisualFX
	{
	public:
		VisualFX() : x(0), y(0), maxR(1), r(1), g(1), b(1) {t.start(1000);};
		VisualFX(float _x, float _y, float _maxR, float _r, float _g, float _b, int msTime) : x(_x), y(_y), maxR(_maxR), r(_r), g(_g), b(_b) {t.start(msTime);};

		bool draw()
		{
			float rad = (maxR - 0.01f)*t.percentComplete() + 0.01f;

			graphics::Circle c;
			c.x1 = x;
			c.y1 = y;
			c.radius = rad;

			glColor4f(r,g,b,1);
			c.draw(0,0,0);
			glColor4f(1,1,1,1);

			return t.done();
		}
	private:
		wok::Timer t;
		float x, y, maxR;
		float r,g,b;
	};

	class Light
	{
	public:
		Light() : x(0), y(0), r(1) {};
		Light(float _x, float _y, float _r) : x(_x), y(_y), r(_r) {};

		void draw(float offsetX, float offsetY)
		{
			graphics::Circle c;
			c.x1 = x;
			c.y1 = y;

			c.radius = r;

			const int M = 15;
			float m = (r - 0.01f) / (float)M;
			float color = 1;
			for(int i = 0; i<=M; ++i)
			{
				color = ((0.001f - 0.5f)/ (float)M) * i + 0.5f;
				glColor4f(color, color, color, 1);
				c.radius = m*i + 0.01f;
				c.draw(offsetX, offsetY, 0);
			}
			glColor4f(1,1,1,1);
		}

		float getX() {return x;}
		float getY() {return y;}
		float getR() {return r;}
	private:
		float x,y,r;
	};
	class Track
		{
		public:
			~Track() {cleanup();}
			void cleanup()
			{
				shapeList.cleanup();
			}
			
			bool loadFromFile(const std::string& filename)
			{
				wok::TextParser text;
	
				text.addDelimiter('=');
				text.addDelimiter(',');
				text.addIgnore(' ');

				std::ifstream f;
				f.open(filename.c_str());
				if(!f.is_open())
					return false;
				char buffer[1024];
				while(!f.eof())
				{
					f.getline(buffer, 1000);
					text.parse(buffer);
					if(text[0] == "Light")
					{
						lights.push_back(Light((float)atof(text[1].c_str()), (float)atof(text[2].c_str()), (float)atof(text[3].c_str())));
					}
					else if(text[0] == "create")
					{
						creationPoints.push_back(math3D::CVector2D((float)atof(text[1].c_str()), (float)atof(text[2].c_str())));
					}
					else if(text[0] == "EOF")
						break;
				}
				f.close();
				shapeList.cleanup();
				nextCP = 0;
				return shapeList.loadFromDataFile(filename);
			}
			
			void draw()
			{	
				if(!lights.empty())
				{
					std::list <Light>::iterator Iter;
					for(Iter = lights.begin(); Iter != lights.end(); ++Iter)
						Iter->draw(0,0);
				}
				shapeList.draw();

				if(!effects.empty())
				{
					std::list <VisualFX>::iterator Iter;
					for(Iter = effects.begin(); Iter != effects.end(); ++Iter)
					{
						if(Iter->draw())
						{
							effects.erase(Iter);
							break;
						}
					}
				}
			}
			
			bool collision(const SRS10::graphics::CollisionCircle& circle)
			{
				return shapeList.checkCollision(circle);
			}

			SRS10::graphics::ShapeCollection& getShapeList() {return shapeList;}

			bool lightCollision(float positionX, float positionY)
			{
				if(lights.empty())
					return false;

				std::list <Light>::iterator Iter;
				for(Iter = lights.begin(); Iter != lights.end(); ++Iter)
				{
					float distance = (positionX - Iter->getX())*(positionX - Iter->getX()) + (positionY - Iter->getY())*(positionY - Iter->getY());
					distance = sqrtf(distance);

					if(distance <= Iter->getR())
						return true;
				}

				return false;
			}

			void addFX(VisualFX& fx)
			{
				effects.push_back(fx);
			}
			
			void getCreationPoint(math3D::CVector2D& p)
			{
				p.x = 0;
				p.y = 0;

				if(nextCP >= creationPoints.size())
					return;
				p = creationPoints[nextCP];
				++nextCP;
				if(nextCP >= creationPoints.size())
					nextCP = 0;
			}
			void getCreationPoint(math3D::CVector2D& p, int index)
			{
				p.x = 0;
				p.y = 0;

				if(index >= creationPoints.size())
					return;
				p = creationPoints[index];
			}
		private:
			SRS10::graphics::ShapeCollection shapeList;
			std::list <Light> lights;
			std::list <VisualFX> effects;
			std::vector <math3D::CVector2D> creationPoints;
			int nextCP;
		};
}
}
#endif